import logging
from contextlib import contextmanager

from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator

import mlflow
from mlflow import MlflowException
from mlflow.entities.trace_info import TraceInfo, TraceState
from mlflow.tracing.provider import get_context_api, get_current_context, get_current_otel_span
from mlflow.tracing.trace_manager import InMemoryTraceManager
from mlflow.tracing.utils import generate_mlflow_trace_id_from_otel_trace_id

_logger = logging.getLogger(__name__)


def get_tracing_context_headers_for_http_request() -> dict[str, str]:
    """
    Get the http request headers that hold information of the tracing context.
    The trace context is serialized as the traceparent header which is defined
    in the W3C TraceContext specification.
    For details, you can refer to
    https://opentelemetry.io/docs/concepts/context-propagation/
    and
    https://www.w3.org/TR/trace-context/#traceparent-header

    Returns:
        The http request headers that hold information of the tracing context.

    Example (client code):

    .. code-block:: python

        import mlflow
        from mlflow.tracing.distributed import get_tracing_context_headers_for_http_request

        with mlflow.start_span("client-root") as client_span:
            # Get the headers that hold information of the tracing context,
            # and send request to remote agent with the headers
            headers = get_tracing_context_headers_for_http_request()
            resp = requests.post(f"{base_url}/remote_agent_handler", headers=headers)

    Example (server handler code):

    .. code-block:: python

        import mlflow
        from flask import Flask, request
        from mlflow.tracing.distributed import set_tracing_context_from_http_request_headers

        app = Flask(__name__)


        @app.post("/agent-handler")
        def handle():
            headers = dict(request.headers)
            with set_tracing_context_from_http_request_headers(headers):
                with mlflow.start_span("server-handler") as span:
                    # call agent ...
                    span.set_attribute("key", "value")
    """
    active_span = mlflow.get_current_active_span()
    if active_span is None:
        _logger.warning(
            "No active span found for fetching the trace context from. Returning an empty header."
        )
    headers = {}
    TraceContextTextMapPropagator().inject(carrier=headers, context=get_current_context())
    return headers


@contextmanager
def set_tracing_context_from_http_request_headers(headers: dict[str, str]):
    """
    Context manager to extract the trace context from the http request headers
    and set the extracted trace context as the current trace context within the
    scope of this context manager.
    The trace context must be serialized as the 'traceparent' header which is defined
    in the W3C TraceContext specification, please see
    :py:func:`mlflow.tracing.distributed.get_tracing_context_headers_for_http_request`
    for how to get the http request headers.

    Args:
        headers: The http request headers to extract the trace context from.

    Example (client code):

    .. code-block:: python

        import mlflow
        from mlflow.tracing.distributed import get_tracing_context_headers_for_http_request

        with mlflow.start_span("client-root") as client_span:
            # Get the headers that hold information of the tracing context,
            # and send request to remote agent with the headers
            headers = get_tracing_context_headers_for_http_request()
            resp = requests.post(f"{base_url}/remote_agent_handler", headers=headers)

    Example (server handler code):

    .. code-block:: python

        import mlflow
        from flask import Flask, request
        from mlflow.tracing.distributed import set_tracing_context_from_http_request_headers

        app = Flask(__name__)


        @app.post("/agent-handler")
        def handle():
            headers = dict(request.headers)
            with set_tracing_context_from_http_request_headers(headers):
                with mlflow.start_span("server-handler") as span:
                    # call agent ...
                    span.set_attribute("key", "value")
    """
    token = None
    otel_trace_id = None
    trace_manager = InMemoryTraceManager.get_instance()
    try:
        headers = dict(headers)

        if "Traceparent" in headers:
            # Note: Some http server framework (e.g. flask) converts http header key
            # first letter to upper case, but `TraceContextTextMapPropagator` can't
            # recognize the key 'Traceparent', so that convert it to lower case.
            traceparent = headers.pop("Traceparent")
            headers["traceparent"] = traceparent

        if "traceparent" not in headers:
            raise MlflowException.invalid_parameter_value(
                "The http request headers do not contain the required key 'traceparent', "
                "please generate the request headers "
                "by 'mlflow.tracing.distributed.get_tracing_context_headers_for_http_request' "
                "API."
            )
        ctx = TraceContextTextMapPropagator().extract(headers)
        token = get_context_api().attach(ctx)

        extracted_span = get_current_otel_span()
        span_context = extracted_span.get_span_context()
        otel_trace_id = span_context.trace_id

        trace_id = generate_mlflow_trace_id_from_otel_trace_id(otel_trace_id)
        dummy_trace_info = TraceInfo(
            trace_id=trace_id,
            trace_location=None,
            request_time=None,
            state=TraceState.IN_PROGRESS,
        )

        trace_manager.register_trace(otel_trace_id, dummy_trace_info, is_remote_trace=True)

        yield
    finally:
        if token is not None:
            get_context_api().detach(token)
        if otel_trace_id is not None:
            trace_manager.pop_trace(otel_trace_id)



from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, AsyncIterable

import numpy as np

from mlflow.entities.gateway_endpoint import FallbackStrategy
from mlflow.exceptions import MlflowException
from mlflow.gateway.base_models import ConfigModel
from mlflow.gateway.config import EndpointConfig
from mlflow.gateway.exceptions import AIGatewayException
from mlflow.gateway.schemas import chat, completions, embeddings
from mlflow.utils.annotations import developer_stable


class PassthroughAction(str, Enum):
    """
    Enum for passthrough endpoint actions.
    """

    OPENAI_CHAT = "openai_chat"
    OPENAI_EMBEDDINGS = "openai_embeddings"
    OPENAI_RESPONSES = "openai_responses"
    ANTHROPIC_MESSAGES = "anthropic_messages"
    GEMINI_GENERATE_CONTENT = "gemini_generate_content"
    GEMINI_STREAM_GENERATE_CONTENT = "gemini_stream_generate_content"


# Mapping of passthrough actions to their gateway API routes
PASSTHROUGH_ROUTES = {
    PassthroughAction.OPENAI_CHAT: "/openai/v1/chat/completions",
    PassthroughAction.OPENAI_EMBEDDINGS: "/openai/v1/embeddings",
    PassthroughAction.OPENAI_RESPONSES: "/openai/v1/responses",
    PassthroughAction.ANTHROPIC_MESSAGES: "/anthropic/v1/messages",
    PassthroughAction.GEMINI_GENERATE_CONTENT: "/gemini/v1beta/models/{endpoint_name}:generateContent",  # noqa: E501
    PassthroughAction.GEMINI_STREAM_GENERATE_CONTENT: "/gemini/v1beta/models/{endpoint_name}:streamGenerateContent",  # noqa: E501
}


@developer_stable
class BaseProvider(ABC):
    """
    Base class for MLflow Gateway providers.
    """

    NAME: str = ""
    SUPPORTED_ROUTE_TYPES: tuple[str, ...]
    CONFIG_TYPE: type[ConfigModel]
    PASSTHROUGH_PROVIDER_PATHS: dict[PassthroughAction, str] = {}

    def __init__(self, config: EndpointConfig):
        if self.NAME == "":
            raise ValueError(
                f"{self.__class__.__name__} is a subclass of BaseProvider and must "
                f"override 'NAME' attribute as a non-empty string."
            )

        if not hasattr(self, "CONFIG_TYPE") or not issubclass(self.CONFIG_TYPE, ConfigModel):
            raise ValueError(
                f"{self.__class__.__name__} is a subclass of BaseProvider and must "
                f"override 'CONFIG_TYPE' attribute as a subclass of ConfigModel."
            )

        self.config = config

    async def chat_stream(
        self, payload: chat.RequestPayload
    ) -> AsyncIterable[chat.StreamResponsePayload]:
        raise AIGatewayException(
            status_code=501,
            detail=f"The chat streaming route is not implemented for {self.NAME} models.",
        )

    async def chat(self, payload: chat.RequestPayload) -> chat.ResponsePayload:
        raise AIGatewayException(
            status_code=501,
            detail=f"The chat route is not implemented for {self.NAME} models.",
        )

    async def completions_stream(
        self, payload: completions.RequestPayload
    ) -> AsyncIterable[completions.StreamResponsePayload]:
        raise AIGatewayException(
            status_code=501,
            detail=f"The completions streaming route is not implemented for {self.NAME} models.",
        )

    async def completions(self, payload: completions.RequestPayload) -> completions.ResponsePayload:
        raise AIGatewayException(
            status_code=501,
            detail=f"The completions route is not implemented for {self.NAME} models.",
        )

    async def embeddings(self, payload: embeddings.RequestPayload) -> embeddings.ResponsePayload:
        raise AIGatewayException(
            status_code=501,
            detail=f"The embeddings route is not implemented for {self.NAME} models.",
        )

    def _validate_passthrough_action(self, action: PassthroughAction) -> str:
        """
        Validates that the passthrough action is supported by this provider
        and returns the provider path.

        Args:
            action: The passthrough action to validate

        Returns:
            The provider path for the action
        """
        provider_path = self.PASSTHROUGH_PROVIDER_PATHS.get(action)
        if provider_path is None:
            route = PASSTHROUGH_ROUTES.get(action, action.value)
            supported_routes = ", ".join(
                f"/gateway{route} (provider_path: {path})"
                for act in self.PASSTHROUGH_PROVIDER_PATHS.keys()
                if (route := PASSTHROUGH_ROUTES.get(act))
                and (path := self.PASSTHROUGH_PROVIDER_PATHS.get(act))
            )
            raise AIGatewayException(
                status_code=400,
                detail=f"Unsupported passthrough endpoint '{route}' for {self.NAME} provider. "
                f"Supported endpoints: {supported_routes}",
            )
        return provider_path

    async def passthrough(
        self,
        action: PassthroughAction,
        payload: dict[str, Any],
        headers: dict[str, str] | None = None,
    ) -> dict[str, Any] | AsyncIterable[bytes]:
        """
        Unified passthrough endpoint for raw API requests.

        Args:
            action: The passthrough action to perform (e.g., OPENAI_CHAT, OPENAI_EMBEDDINGS)
            payload: Raw request payload in the format expected by the target API
            headers: Optional HTTP headers from client request to propagate

        Returns:
            Raw response from the target API, optionally as an async iterable for streaming

        Raises:
            AIGatewayException: If the passthrough action is not implemented for this provider
        """
        route = PASSTHROUGH_ROUTES.get(action)
        raise AIGatewayException(
            status_code=501,
            detail=f"The passthrough route '{route}' is not implemented for {self.NAME} models.",
        )

    @staticmethod
    def check_for_model_field(payload):
        if "model" in payload:
            raise AIGatewayException(
                status_code=422,
                detail="The parameter 'model' is not permitted to be passed. The route being "
                "queried already defines a model instance.",
            )


class TrafficRouteProvider(BaseProvider):
    """
    A provider that split traffic and forward to multiple providers
    """

    NAME: str = "TrafficRoute"

    def __init__(
        self,
        configs: list[EndpointConfig],
        traffic_splits: list[int],
        routing_strategy: str,
    ):
        from mlflow.gateway.providers import get_provider

        if len(configs) != len(traffic_splits):
            raise MlflowException.invalid_parameter_value(
                "'configs' and 'traffic_splits' should have the same length."
            )

        if routing_strategy != "TRAFFIC_SPLIT":
            raise MlflowException.invalid_parameter_value(
                "'routing_strategy' must be 'TRAFFIC_SPLIT'."
            )

        self._providers = [get_provider(config.model.provider)(config) for config in configs]
        # Normalize the weights to sum to 1
        self._weights = np.array(traffic_splits, dtype=np.float32) / np.sum(traffic_splits)
        self._indices = np.arange(len(self._providers))

    def _get_provider(self):
        chosen_index = np.random.choice(self._indices, p=self._weights)
        return self._providers[chosen_index]

    async def chat_stream(
        self, payload: chat.RequestPayload
    ) -> AsyncIterable[chat.StreamResponsePayload]:
        prov = self._get_provider()
        async for i in prov.chat_stream(payload):
            yield i

    async def chat(self, payload: chat.RequestPayload) -> chat.ResponsePayload:
        prov = self._get_provider()
        return await prov.chat(payload)

    async def completions_stream(
        self, payload: completions.RequestPayload
    ) -> AsyncIterable[completions.StreamResponsePayload]:
        prov = self._get_provider()
        async for i in prov.completions_stream(payload):
            yield i

    async def completions(self, payload: completions.RequestPayload) -> completions.ResponsePayload:
        prov = self._get_provider()
        return await prov.completions(payload)

    async def embeddings(self, payload: embeddings.RequestPayload) -> embeddings.ResponsePayload:
        prov = self._get_provider()
        return await prov.embeddings(payload)

    async def passthrough(
        self,
        action: PassthroughAction,
        payload: dict[str, Any],
        headers: dict[str, str] | None = None,
    ) -> dict[str, Any] | AsyncIterable[bytes]:
        prov = self._get_provider()
        return await prov.passthrough(action, payload, headers)


class FallbackProvider(BaseProvider):
    """
    A provider that implements fallback routing across multiple providers.

    Attempts to call providers in order until one succeeds or max_attempts is reached.
    """

    NAME: str = "Fallback"

    def __init__(
        self,
        providers: list[BaseProvider],
        strategy: FallbackStrategy | None = None,
        max_attempts: int | None = None,
    ):
        if not providers:
            raise MlflowException.invalid_parameter_value(
                "'providers' must contain at least one provider."
            )

        self._providers = providers

        max_attempts = max_attempts if max_attempts is not None else len(self._providers)
        self._max_attempts = min(max_attempts, len(self._providers))
        self._strategy = strategy

    async def _execute_with_fallback(self, method_name: str, *args, **kwargs):
        """
        Execute a method on providers with fallback logic.

        Args:
            method_name: Name of the method to call on each provider
            *args: Positional arguments to pass to the method
            **kwargs: Keyword arguments to pass to the method

        Returns:
            Result from the first successful provider call

        Raises:
            AIGatewayException: If all fallback attempts fail, with status code
                propagated from the last exception if it was an AIGatewayException
                or HTTPException
        """
        from fastapi import HTTPException

        last_error = None

        for attempt, provider in enumerate(self._providers[: self._max_attempts], 1):
            try:
                method = getattr(provider, method_name)
                return await method(*args, **kwargs)
            except Exception as e:
                last_error = e
                if attempt < self._max_attempts:
                    continue
                break

        # Propagate HTTP status code from the last exception if available
        status_code = 500
        if isinstance(last_error, (AIGatewayException, HTTPException)):
            status_code = last_error.status_code

        raise AIGatewayException(
            status_code=status_code,
            detail=f"All {self._max_attempts} fallback attempts failed. Last error: {last_error!s}",
        )

    async def _execute_stream_with_fallback(self, method_name: str, *args, **kwargs):
        """
        Execute a streaming method on providers with fallback logic.

        Args:
            method_name: Name of the streaming method to call on each provider
            *args: Positional arguments to pass to the method
            **kwargs: Keyword arguments to pass to the method

        Yields:
            Stream chunks from the first successful provider call

        Raises:
            AIGatewayException: If all fallback attempts fail, with status code
                propagated from the last exception if it was an AIGatewayException
                or HTTPException
        """
        from fastapi import HTTPException

        last_error = None

        for attempt, provider in enumerate(self._providers[: self._max_attempts], 1):
            try:
                method = getattr(provider, method_name)
                async for chunk in method(*args, **kwargs):
                    yield chunk
                return
            except Exception as e:
                last_error = e
                if attempt < self._max_attempts:
                    continue
                break

        # Propagate HTTP status code from the last exception if available
        status_code = 500
        if isinstance(last_error, (AIGatewayException, HTTPException)):
            status_code = last_error.status_code

        raise AIGatewayException(
            status_code=status_code,
            detail=f"All {self._max_attempts} fallback attempts failed. Last error: {last_error!s}",
        )

    async def chat_stream(
        self, payload: chat.RequestPayload
    ) -> AsyncIterable[chat.StreamResponsePayload]:
        async for chunk in self._execute_stream_with_fallback("chat_stream", payload):
            yield chunk

    async def chat(self, payload: chat.RequestPayload) -> chat.ResponsePayload:
        return await self._execute_with_fallback("chat", payload)

    async def completions_stream(
        self, payload: completions.RequestPayload
    ) -> AsyncIterable[completions.StreamResponsePayload]:
        async for chunk in self._execute_stream_with_fallback("completions_stream", payload):
            yield chunk

    async def completions(self, payload: completions.RequestPayload) -> completions.ResponsePayload:
        return await self._execute_with_fallback("completions", payload)

    async def embeddings(self, payload: embeddings.RequestPayload) -> embeddings.ResponsePayload:
        return await self._execute_with_fallback("embeddings", payload)

    async def passthrough(
        self,
        action: PassthroughAction,
        payload: dict[str, Any],
        headers: dict[str, str] | None = None,
    ) -> dict[str, Any] | AsyncIterable[bytes]:
        return await self._execute_with_fallback("passthrough", action, payload, headers)


class ProviderAdapter(ABC):
    @classmethod
    @abstractmethod
    def model_to_embeddings(cls, resp, config): ...

    @classmethod
    @abstractmethod
    def model_to_completions(cls, resp, config): ...

    @classmethod
    def model_to_completions_streaming(cls, resp, config):
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def completions_to_model(cls, payload, config): ...

    @classmethod
    def completions_streaming_to_model(cls, payload, config):
        raise NotImplementedError

    @classmethod
    def model_to_chat(cls, resp, config):
        raise NotImplementedError

    @classmethod
    def model_to_chat_streaming(cls, resp, config):
        raise NotImplementedError

    @classmethod
    def chat_to_model(cls, payload, config):
        raise NotImplementedError

    @classmethod
    def chat_streaming_to_model(cls, payload, config):
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def embeddings_to_model(cls, payload, config): ...

    @classmethod
    def check_keys_against_mapping(cls, mapping, payload):
        for k1, k2 in mapping.items():
            if k2 in payload:
                raise AIGatewayException(
                    status_code=400, detail=f"Invalid parameter {k2}. Use {k1} instead."
                )



from __future__ import annotations

import json
import logging
import uuid
from concurrent.futures import ThreadPoolExecutor, as_completed
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Callable

import pydantic

import mlflow
from mlflow.environment_variables import MLFLOW_GENAI_EVAL_MAX_WORKERS
from mlflow.exceptions import MlflowException
from mlflow.genai.datasets import EvaluationDataset
from mlflow.genai.judges.adapters.databricks_managed_judge_adapter import (
    call_chat_completions,
    create_litellm_message_from_databricks_response,
    serialize_messages_to_databricks_prompts,
)
from mlflow.genai.judges.adapters.databricks_serving_endpoint_adapter import (
    _invoke_databricks_serving_endpoint,
)
from mlflow.genai.judges.constants import (
    _DATABRICKS_AGENTIC_JUDGE_MODEL,
    _DATABRICKS_DEFAULT_JUDGE_MODEL,
)
from mlflow.genai.judges.utils import get_default_model
from mlflow.genai.simulators.prompts import (
    CHECK_GOAL_PROMPT,
    DEFAULT_PERSONA,
    FOLLOWUP_USER_PROMPT,
    INITIAL_USER_PROMPT,
)
from mlflow.genai.utils.trace_utils import parse_outputs_to_str
from mlflow.telemetry.events import SimulateConversationEvent
from mlflow.telemetry.track import record_usage_event
from mlflow.tracing.constant import TraceMetadataKey
from mlflow.utils.annotations import experimental
from mlflow.utils.docstring_utils import format_docstring

try:
    from tqdm.auto import tqdm
except ImportError:
    tqdm = None

if TYPE_CHECKING:
    from pandas import DataFrame

    from mlflow.types.llm import ChatMessage

_logger = logging.getLogger(__name__)

_MAX_METADATA_LENGTH = 250
_EXPECTED_TEST_CASE_KEYS = {"goal", "persona", "context"}
_REQUIRED_TEST_CASE_KEYS = {"goal"}

_MODEL_API_DOC = {
    "model": """Model to use for generating user messages. Must be one of:

* `"databricks"` - Uses the Databricks managed LLM endpoint
* `"databricks:/<endpoint-name>"` - Uses a Databricks model serving endpoint \
(e.g., `"databricks:/databricks-claude-sonnet-4-5"`)
* `"<provider>:/<model-name>"` - Uses LiteLLM (e.g., `"openai:/gpt-4.1-mini"`, \
`"anthropic:/claude-3.5-sonnet-20240620"`)

MLflow natively supports `["openai", "anthropic", "bedrock", "mistral"]`, and more \
providers are supported through `LiteLLM <https://docs.litellm.ai/docs/providers>`_.

Default model depends on the tracking URI setup:

* Databricks: `"databricks"`
* Otherwise: `"openai:/gpt-4.1-mini"`
""",
}


class GoalCheckResult(pydantic.BaseModel):
    """Structured output for goal achievement check."""

    rationale: str = pydantic.Field(
        description="Reason for the assessment explaining whether the goal has been achieved"
    )
    result: str = pydantic.Field(description="'yes' if goal achieved, 'no' otherwise")


@contextmanager
def _suppress_tracing_logging():
    # Suppress INFO logs when flushing traces from the async trace export queue.
    async_logger = logging.getLogger("mlflow.tracing.export.async_export_queue")
    # Suppress WARNING logs when the tracing provider is automatically traced, but used in a
    # tracing-disabled context (e.g., generating user messages).
    fluent_logger = logging.getLogger("mlflow.tracing.fluent")
    original_async_level = async_logger.level
    original_fluent_level = fluent_logger.level

    async_logger.setLevel(logging.WARNING)
    fluent_logger.setLevel(logging.ERROR)

    try:
        yield
    finally:
        async_logger.setLevel(original_async_level)
        fluent_logger.setLevel(original_fluent_level)


@contextmanager
def _delete_trace_if_created():
    """Delete any trace created within this context to avoid polluting user traces."""
    trace_id_before = mlflow.get_last_active_trace_id(thread_local=True)
    try:
        yield
    finally:
        trace_id_after = mlflow.get_last_active_trace_id(thread_local=True)
        if trace_id_after and trace_id_after != trace_id_before:
            try:
                mlflow.delete_trace(trace_id_after)
            except Exception as e:
                _logger.debug(f"Failed to delete trace {trace_id_after}: {e}")


# TODO: Refactor judges adapters to support returning raw responses, then use them here
#       instead of reimplementing the invocation logic.
def _invoke_model_without_tracing(
    model_uri: str,
    messages: list["ChatMessage"],
    num_retries: int = 3,
    inference_params: dict[str, Any] | None = None,
) -> str:
    """
    Invoke a model without tracing. This method will delete the last trace created by the
    invocation, if any.
    """
    import litellm

    from mlflow.metrics.genai.model_utils import _parse_model_uri

    with _delete_trace_if_created():
        # Use Databricks managed endpoint with agentic model for the default "databricks" URI
        if model_uri == _DATABRICKS_DEFAULT_JUDGE_MODEL:
            user_prompt, system_prompt = serialize_messages_to_databricks_prompts(messages)

            result = call_chat_completions(
                user_prompt=user_prompt,
                # NB: We cannot use an empty system prompt here so we use a period.
                system_prompt=system_prompt or ".",
                model=_DATABRICKS_AGENTIC_JUDGE_MODEL,
            )
            if getattr(result, "error_code", None):
                raise MlflowException(
                    f"Failed to get chat completions result from Databricks managed endpoint: "
                    f"[{result.error_code}] {result.error_message}"
                )

            output_json = result.output_json
            if not output_json:
                raise MlflowException("Empty response from Databricks managed endpoint")

            parsed_json = json.loads(output_json) if isinstance(output_json, str) else output_json
            return create_litellm_message_from_databricks_response(parsed_json).content

        provider, model_name = _parse_model_uri(model_uri)

        # Use Databricks serving endpoint for databricks:/<endpoint> URIs
        if provider in {"databricks", "endpoints"}:
            output = _invoke_databricks_serving_endpoint(
                model_name=model_name,
                prompt=messages,
                num_retries=num_retries,
                inference_params=inference_params,
            )
            return output.response

        # Use LiteLLM for other providers
        litellm_messages = [litellm.Message(role=msg.role, content=msg.content) for msg in messages]

        kwargs = {
            "model": f"{provider}/{model_name}",
            "messages": litellm_messages,
            "max_retries": num_retries,
        }
        if inference_params:
            kwargs.update(inference_params)

        response = litellm.completion(**kwargs)
        return response.choices[0].message.content


def _get_last_response(conversation_history: list[dict[str, Any]]) -> str | None:
    if not conversation_history:
        return None

    last_msg = conversation_history[-1]
    content = last_msg.get("content")
    if isinstance(content, str) and content:
        return content

    result = parse_outputs_to_str(last_msg)
    if result and result.strip():
        return result

    return str(last_msg)


def _format_history(history: list[dict[str, Any]]) -> str | None:
    if not history:
        return None
    formatted = []
    for msg in history:
        role = msg.get("role") or "unknown"
        content = msg.get("content") or ""
        formatted.append(f"{role}: {content}")
    return "\n".join(formatted)


@format_docstring(_MODEL_API_DOC)
@experimental(version="3.9.0")
class SimulatedUserAgent:
    """
    An LLM-powered agent that simulates user behavior in conversations.

    The agent generates realistic user messages based on a specified goal and persona,
    enabling automated testing of conversational AI systems.

    Args:
        goal: The objective the simulated user is trying to achieve in the conversation.
        persona: Description of the user's personality and background. If None, uses a
            default helpful user persona.
        model: {{ model }}
        **inference_params: Additional parameters passed to the LLM (e.g., temperature).
    """

    def __init__(
        self,
        goal: str,
        persona: str | None = None,
        model: str | None = None,
        **inference_params,
    ):
        self.goal = goal
        self.persona = persona or DEFAULT_PERSONA
        self.model = model or get_default_model()
        self.inference_params = inference_params

    def generate_message(
        self,
        conversation_history: list[dict[str, Any]],
        turn: int = 0,
    ) -> str:
        from mlflow.types.llm import ChatMessage

        if turn == 0:
            prompt = INITIAL_USER_PROMPT.format(persona=self.persona, goal=self.goal)
        else:
            last_response = _get_last_response(conversation_history)
            history_str = _format_history(conversation_history[:-1])
            prompt = FOLLOWUP_USER_PROMPT.format(
                persona=self.persona,
                goal=self.goal,
                conversation_history=history_str if history_str is not None else "",
                last_response=last_response if last_response is not None else "",
            )

        messages = [ChatMessage(role="user", content=prompt)]
        return _invoke_model_without_tracing(
            model_uri=self.model,
            messages=messages,
            num_retries=3,
            inference_params=self.inference_params,
        )


@format_docstring(_MODEL_API_DOC)
@experimental(version="3.9.0")
class ConversationSimulator:
    """
    Generates multi-turn conversations by simulating user interactions with a target agent.

    The simulator creates a simulated user agent that interacts with your agent's predict function.
    Each conversation is traced in MLflow, allowing you to evaluate how your agent handles
    various user goals and personas.

    The predict function passed to the simulator must follow the OpenAI Responses API format
    (https://platform.openai.com/docs/api-reference/responses):

    - It must accept an ``input`` parameter containing the conversation history
      as a list of message dictionaries (e.g., ``[{"role": "user", "content": "..."}]``)
    - It may accept additional keyword arguments from the test case's ``context`` field
    - It should return a response (the assistant's message content will be extracted)

    Args:
        test_cases: List of test case dicts, a DataFrame, or an EvaluationDataset,
            with the following fields:

            - "goal": Describing what the simulated user wants to achieve.
            - "persona" (optional): Custom persona for the simulated user.
            - "context" (optional): Dict of additional kwargs to pass to predict_fn.

        max_turns: Maximum number of conversation turns before stopping. Default is 10.
        user_model: {{ model }}
        **user_llm_params: Additional parameters passed to the simulated user's LLM calls.

    Example:
        .. code-block:: python

            import mlflow
            from mlflow.genai.simulators import ConversationSimulator
            from mlflow.genai.scorers import ConversationalSafety, Safety


            def predict_fn(input: list[dict], **kwargs) -> dict:
                response = client.chat.completions.create(
                    model="gpt-4o-mini",
                    messages=input,
                )
                return response


            # Each test case requires a "goal". "persona" and "context" are optional.
            simulator = ConversationSimulator(
                test_cases=[
                    {"goal": "Learn about MLflow tracking"},
                    {"goal": "Debug deployment issue", "persona": "A data scientist"},
                    {
                        "goal": "Set up model registry",
                        "persona": "A beginner",
                        "context": {"user_id": "123"},
                    },
                ],
                max_turns=5,
            )

            mlflow.genai.evaluate(
                data=simulator,
                predict_fn=predict_fn,
                scorers=[ConversationalSafety(), Safety()],
            )
    """

    def __init__(
        self,
        test_cases: list[dict[str, Any]] | "DataFrame" | EvaluationDataset,
        max_turns: int = 10,
        user_model: str | None = None,
        **user_llm_params,
    ):
        self.test_cases = self._normalize_test_cases(test_cases)
        self._validate_test_cases()
        self.max_turns = max_turns
        self.user_model = user_model or get_default_model()
        self.user_llm_params = user_llm_params

    def _normalize_test_cases(
        self, test_cases: list[dict[str, Any]] | "DataFrame" | EvaluationDataset
    ) -> list[dict[str, Any]]:
        from pandas import DataFrame

        if isinstance(test_cases, EvaluationDataset):
            records = test_cases.to_df()["inputs"].to_list()
            if not records or not (records[0].keys() & _REQUIRED_TEST_CASE_KEYS):
                raise ValueError(
                    "EvaluationDataset passed to ConversationSimulator must contain "
                    "conversational test cases with a 'goal' field in the 'inputs' column"
                )
            return records

        if isinstance(test_cases, DataFrame):
            return test_cases.to_dict("records")

        return test_cases

    def _validate_test_cases(self) -> None:
        if not self.test_cases:
            raise ValueError("test_cases cannot be empty")

        missing_goal_indices = [
            i for i, test_case in enumerate(self.test_cases) if not test_case.get("goal")
        ]
        if missing_goal_indices:
            raise ValueError(f"Test cases at indices {missing_goal_indices} must have 'goal' field")

        indices_with_extra_keys = [
            i
            for i, test_case in enumerate(self.test_cases)
            if set(test_case.keys()) - _EXPECTED_TEST_CASE_KEYS
        ]
        if indices_with_extra_keys:
            _logger.warning(
                f"Test cases at indices {indices_with_extra_keys} contain unexpected keys "
                f"which will be ignored. Expected keys: {_EXPECTED_TEST_CASE_KEYS}."
            )

    @record_usage_event(SimulateConversationEvent)
    def _simulate(self, predict_fn: Callable[..., dict[str, Any]]) -> list[list[str]]:
        num_test_cases = len(self.test_cases)
        all_trace_ids: list[list[str]] = [[] for _ in range(num_test_cases)]
        max_workers = min(num_test_cases, MLFLOW_GENAI_EVAL_MAX_WORKERS.get())

        progress_bar = (
            tqdm(
                total=num_test_cases,
                desc="Simulating conversations",
                unit="conversation",
            )
            if tqdm
            else None
        )

        with (
            _suppress_tracing_logging(),
            ThreadPoolExecutor(
                max_workers=max_workers,
                thread_name_prefix="MlflowConversationSimulator",
            ) as executor,
        ):
            futures = {
                executor.submit(self._run_conversation, test_case, predict_fn): i
                for i, test_case in enumerate(self.test_cases)
            }
            try:
                for future in as_completed(futures):
                    idx = futures[future]
                    try:
                        all_trace_ids[idx] = future.result()
                    except Exception as e:
                        _logger.error(
                            f"Failed to run conversation for test case "
                            f"{self.test_cases[idx].get('goal')}: {e}"
                        )
                    if progress_bar:
                        progress_bar.update(1)
            finally:
                if progress_bar:
                    progress_bar.close()

        return all_trace_ids

    def _run_conversation(
        self, test_case: dict[str, Any], predict_fn: Callable[..., dict[str, Any]]
    ) -> list[str]:
        goal = test_case["goal"]
        persona = test_case.get("persona")
        context = test_case.get("context", {})
        trace_session_id = f"sim-{uuid.uuid4().hex[:16]}"

        user_agent = SimulatedUserAgent(
            goal=goal,
            persona=persona,
            model=self.user_model,
            **self.user_llm_params,
        )

        conversation_history: list[dict[str, Any]] = []
        trace_ids: list[str] = []

        for turn in range(self.max_turns):
            try:
                # Generate a user message using the simulated user agent.
                user_message_content = user_agent.generate_message(conversation_history, turn)
                user_message = {"role": "user", "content": user_message_content}
                conversation_history.append(user_message)

                # Invoke the predict_fn with the user message and context.
                response, trace_id = self._invoke_predict_fn(
                    predict_fn=predict_fn,
                    input_messages=conversation_history,
                    trace_session_id=trace_session_id,
                    goal=goal,
                    persona=persona,
                    turn=turn,
                    **context,
                )

                if trace_id:
                    trace_ids.append(trace_id)

                # Parse the assistant's response and check if the goal has been achieved.
                assistant_content = parse_outputs_to_str(response)
                if not assistant_content or not assistant_content.strip():
                    _logger.debug(f"Stopping conversation: empty response at turn {turn}")
                    break
                conversation_history.append({"role": "assistant", "content": assistant_content})
                if self._check_goal_achieved(conversation_history, assistant_content, goal):
                    _logger.debug(f"Stopping conversation: goal achieved at turn {turn}")
                    break

            except Exception as e:
                _logger.error(f"Error during turn {turn}: {e}", exc_info=True)
                break

        return trace_ids

    def _invoke_predict_fn(
        self,
        predict_fn: Callable[..., dict[str, Any]],
        input_messages: list[dict[str, Any]],
        trace_session_id: str,
        goal: str,
        persona: str | None,
        turn: int,
        **context,
    ) -> tuple[dict[str, Any], str | None]:
        # NB: We trace the predict_fn call to add session and simulation metadata to the trace.
        #     This adds a new root span to the trace, with the same inputs and outputs as the
        #     predict_fn call. The goal/persona/turn metadata is used for trace comparison UI
        #     since message content may differ between simulation runs.
        @mlflow.trace(name=f"simulation_turn_{turn}", span_type="CHAIN")
        def traced_predict(input: list[dict[str, Any]], **ctx):
            mlflow.update_current_trace(
                metadata={
                    TraceMetadataKey.TRACE_SESSION: trace_session_id,
                    "mlflow.simulation.goal": goal[:_MAX_METADATA_LENGTH],
                    "mlflow.simulation.persona": (persona or DEFAULT_PERSONA)[
                        :_MAX_METADATA_LENGTH
                    ],
                    "mlflow.simulation.turn": str(turn),
                },
            )
            return predict_fn(input=input, **ctx)

        response = traced_predict(input=input_messages, **context)
        trace_id = mlflow.get_last_active_trace_id(thread_local=True)
        return response, trace_id

    def _check_goal_achieved(
        self,
        conversation_history: list[dict[str, Any]],
        last_response: str,
        goal: str,
    ) -> bool:
        from mlflow.types.llm import ChatMessage

        history_str = _format_history(conversation_history)
        eval_prompt = CHECK_GOAL_PROMPT.format(
            goal=goal,
            conversation_history=history_str if history_str is not None else "",
            last_response=last_response,
        )
        messages = [ChatMessage(role="user", content=eval_prompt)]

        try:
            text_result = _invoke_model_without_tracing(
                model_uri=self.user_model,
                messages=messages,
                num_retries=3,
                inference_params={"temperature": 0.0},
            )
            result = GoalCheckResult.model_validate_json(text_result)
            return result.result.strip().lower() == "yes"
        except pydantic.ValidationError:
            _logger.warning(f"Could not parse response for goal achievement check: {text_result}")
            return False
        except Exception as e:
            _logger.warning(f"Goal achievement check failed: {e}")
            return False



from sqlalchemy.exc import IntegrityError, MultipleResultsFound, NoResultFound
from sqlalchemy.orm import sessionmaker
from werkzeug.security import check_password_hash, generate_password_hash

from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import (
    INVALID_STATE,
    RESOURCE_ALREADY_EXISTS,
    RESOURCE_DOES_NOT_EXIST,
)
from mlflow.server.auth.db import utils as dbutils
from mlflow.server.auth.db.models import (
    SqlExperimentPermission,
    SqlGatewayEndpointPermission,
    SqlGatewayModelDefinitionPermission,
    SqlGatewaySecretPermission,
    SqlRegisteredModelPermission,
    SqlScorerPermission,
    SqlUser,
)
from mlflow.server.auth.entities import (
    ExperimentPermission,
    GatewayEndpointPermission,
    GatewayModelDefinitionPermission,
    GatewaySecretPermission,
    RegisteredModelPermission,
    ScorerPermission,
    User,
)
from mlflow.server.auth.permissions import _validate_permission
from mlflow.store.db.utils import _get_managed_session_maker, create_sqlalchemy_engine_with_retry
from mlflow.utils.uri import extract_db_type_from_uri
from mlflow.utils.validation import _validate_password, _validate_username


class SqlAlchemyStore:
    def init_db(self, db_uri):
        self.db_uri = db_uri
        self.db_type = extract_db_type_from_uri(db_uri)
        self.engine = create_sqlalchemy_engine_with_retry(db_uri)
        dbutils.migrate_if_needed(self.engine, "head")
        SessionMaker = sessionmaker(bind=self.engine)
        self.ManagedSessionMaker = _get_managed_session_maker(SessionMaker, self.db_type)

    def authenticate_user(self, username: str, password: str) -> bool:
        with self.ManagedSessionMaker() as session:
            try:
                user = self._get_user(session, username)
                return check_password_hash(user.password_hash, password)
            except MlflowException:
                return False

    def create_user(self, username: str, password: str, is_admin: bool = False) -> User:
        _validate_username(username)
        _validate_password(password)
        pwhash = generate_password_hash(password)
        with self.ManagedSessionMaker() as session:
            try:
                user = SqlUser(username=username, password_hash=pwhash, is_admin=is_admin)
                session.add(user)
                session.flush()
                return user.to_mlflow_entity()
            except IntegrityError as e:
                raise MlflowException(
                    f"User (username={username}) already exists. Error: {e}",
                    RESOURCE_ALREADY_EXISTS,
                ) from e

    @staticmethod
    def _get_user(session, username: str) -> SqlUser:
        try:
            return session.query(SqlUser).filter(SqlUser.username == username).one()
        except NoResultFound:
            raise MlflowException(
                f"User with username={username} not found",
                RESOURCE_DOES_NOT_EXIST,
            )
        except MultipleResultsFound:
            raise MlflowException(
                f"Found multiple users with username={username}",
                INVALID_STATE,
            )

    def has_user(self, username: str) -> bool:
        with self.ManagedSessionMaker() as session:
            return session.query(SqlUser).filter(SqlUser.username == username).first() is not None

    def get_user(self, username: str) -> User:
        with self.ManagedSessionMaker() as session:
            return self._get_user(session, username).to_mlflow_entity()

    def list_users(self) -> list[User]:
        with self.ManagedSessionMaker() as session:
            users = session.query(SqlUser).all()
            return [u.to_mlflow_entity() for u in users]

    def update_user(
        self, username: str, password: str | None = None, is_admin: bool | None = None
    ) -> User:
        with self.ManagedSessionMaker() as session:
            user = self._get_user(session, username)
            if password is not None:
                pwhash = generate_password_hash(password)
                user.password_hash = pwhash
            if is_admin is not None:
                user.is_admin = is_admin
            return user.to_mlflow_entity()

    def delete_user(self, username: str):
        with self.ManagedSessionMaker() as session:
            user = self._get_user(session, username)
            session.delete(user)

    def create_experiment_permission(
        self, experiment_id: str, username: str, permission: str
    ) -> ExperimentPermission:
        _validate_permission(permission)
        with self.ManagedSessionMaker() as session:
            try:
                user = self._get_user(session, username=username)
                perm = SqlExperimentPermission(
                    experiment_id=experiment_id, user_id=user.id, permission=permission
                )
                session.add(perm)
                session.flush()
                return perm.to_mlflow_entity()
            except IntegrityError as e:
                raise MlflowException(
                    f"Experiment permission (experiment_id={experiment_id}, username={username}) "
                    f"already exists. Error: {e}",
                    RESOURCE_ALREADY_EXISTS,
                )

    def _get_experiment_permission(
        self, session, experiment_id: str, username: str
    ) -> SqlExperimentPermission:
        try:
            user = self._get_user(session, username=username)
            return (
                session.query(SqlExperimentPermission)
                .filter(
                    SqlExperimentPermission.experiment_id == experiment_id,
                    SqlExperimentPermission.user_id == user.id,
                )
                .one()
            )
        except NoResultFound:
            raise MlflowException(
                f"Experiment permission with experiment_id={experiment_id} and "
                f"username={username} not found",
                RESOURCE_DOES_NOT_EXIST,
            )
        except MultipleResultsFound:
            raise MlflowException(
                f"Found multiple experiment permissions with experiment_id={experiment_id} "
                f"and username={username}",
                INVALID_STATE,
            )

    def get_experiment_permission(self, experiment_id: str, username: str) -> ExperimentPermission:
        with self.ManagedSessionMaker() as session:
            return self._get_experiment_permission(
                session, experiment_id, username
            ).to_mlflow_entity()

    def list_experiment_permissions(self, username: str) -> list[ExperimentPermission]:
        with self.ManagedSessionMaker() as session:
            user = self._get_user(session, username=username)
            perms = (
                session.query(SqlExperimentPermission)
                .filter(SqlExperimentPermission.user_id == user.id)
                .all()
            )
            return [p.to_mlflow_entity() for p in perms]

    def update_experiment_permission(
        self, experiment_id: str, username: str, permission: str
    ) -> ExperimentPermission:
        _validate_permission(permission)
        with self.ManagedSessionMaker() as session:
            perm = self._get_experiment_permission(session, experiment_id, username)
            perm.permission = permission
            return perm.to_mlflow_entity()

    def delete_experiment_permission(self, experiment_id: str, username: str):
        with self.ManagedSessionMaker() as session:
            perm = self._get_experiment_permission(session, experiment_id, username)
            session.delete(perm)

    def create_registered_model_permission(
        self, name: str, username: str, permission: str
    ) -> RegisteredModelPermission:
        _validate_permission(permission)
        with self.ManagedSessionMaker() as session:
            try:
                user = self._get_user(session, username=username)
                perm = SqlRegisteredModelPermission(
                    name=name, user_id=user.id, permission=permission
                )
                session.add(perm)
                session.flush()
                return perm.to_mlflow_entity()
            except IntegrityError as e:
                raise MlflowException(
                    f"Registered model permission (name={name}, username={username}) "
                    f"already exists. Error: {e}",
                    RESOURCE_ALREADY_EXISTS,
                )

    def _get_registered_model_permission(
        self, session, name: str, username: str
    ) -> SqlRegisteredModelPermission:
        try:
            user = self._get_user(session, username=username)
            return (
                session.query(SqlRegisteredModelPermission)
                .filter(
                    SqlRegisteredModelPermission.name == name,
                    SqlRegisteredModelPermission.user_id == user.id,
                )
                .one()
            )
        except NoResultFound:
            raise MlflowException(
                f"Registered model permission with name={name} and username={username} not found",
                RESOURCE_DOES_NOT_EXIST,
            )
        except MultipleResultsFound:
            raise MlflowException(
                f"Found multiple registered model permissions with name={name} "
                f"and username={username}",
                INVALID_STATE,
            )

    def get_registered_model_permission(
        self, name: str, username: str
    ) -> RegisteredModelPermission:
        with self.ManagedSessionMaker() as session:
            return self._get_registered_model_permission(session, name, username).to_mlflow_entity()

    def list_registered_model_permissions(self, username: str) -> list[RegisteredModelPermission]:
        with self.ManagedSessionMaker() as session:
            user = self._get_user(session, username=username)
            perms = (
                session.query(SqlRegisteredModelPermission)
                .filter(SqlRegisteredModelPermission.user_id == user.id)
                .all()
            )
            return [p.to_mlflow_entity() for p in perms]

    def update_registered_model_permission(
        self, name: str, username: str, permission: str
    ) -> RegisteredModelPermission:
        _validate_permission(permission)
        with self.ManagedSessionMaker() as session:
            perm = self._get_registered_model_permission(session, name, username)
            perm.permission = permission
            return perm.to_mlflow_entity()

    def delete_registered_model_permission(self, name: str, username: str):
        with self.ManagedSessionMaker() as session:
            perm = self._get_registered_model_permission(session, name, username)
            session.delete(perm)

    def rename_registered_model_permissions(self, old_name: str, new_name: str):
        with self.ManagedSessionMaker() as session:
            perms = (
                session.query(SqlRegisteredModelPermission)
                .filter(SqlRegisteredModelPermission.name == old_name)
                .all()
            )
            for perm in perms:
                perm.name = new_name

    def create_scorer_permission(
        self, experiment_id: str, scorer_name: str, username: str, permission: str
    ) -> ScorerPermission:
        _validate_permission(permission)
        with self.ManagedSessionMaker() as session:
            try:
                user = self._get_user(session, username=username)
                perm = SqlScorerPermission(
                    experiment_id=experiment_id,
                    scorer_name=scorer_name,
                    user_id=user.id,
                    permission=permission,
                )
                session.add(perm)
                session.flush()
                return perm.to_mlflow_entity()
            except IntegrityError as e:
                raise MlflowException(
                    f"Scorer permission (experiment_id={experiment_id}, scorer_name={scorer_name}, "
                    f"username={username}) already exists. Error: {e}",
                    RESOURCE_ALREADY_EXISTS,
                ) from e

    def _get_scorer_permission(
        self, session, experiment_id: str, scorer_name: str, username: str
    ) -> SqlScorerPermission:
        try:
            user = self._get_user(session, username=username)
            return (
                session.query(SqlScorerPermission)
                .filter(
                    SqlScorerPermission.experiment_id == experiment_id,
                    SqlScorerPermission.scorer_name == scorer_name,
                    SqlScorerPermission.user_id == user.id,
                )
                .one()
            )
        except NoResultFound:
            raise MlflowException(
                f"Scorer permission with experiment_id={experiment_id}, "
                f"scorer_name={scorer_name}, and username={username} not found",
                RESOURCE_DOES_NOT_EXIST,
            )
        except MultipleResultsFound:
            raise MlflowException(
                f"Found multiple scorer permissions with experiment_id={experiment_id}, "
                f"scorer_name={scorer_name}, and username={username}",
                INVALID_STATE,
            )

    def get_scorer_permission(
        self, experiment_id: str, scorer_name: str, username: str
    ) -> ScorerPermission:
        with self.ManagedSessionMaker() as session:
            return self._get_scorer_permission(
                session, experiment_id, scorer_name, username
            ).to_mlflow_entity()

    def list_scorer_permissions(self, username: str) -> list[ScorerPermission]:
        with self.ManagedSessionMaker() as session:
            user = self._get_user(session, username=username)
            perms = (
                session.query(SqlScorerPermission)
                .filter(SqlScorerPermission.user_id == user.id)
                .all()
            )
            return [p.to_mlflow_entity() for p in perms]

    def update_scorer_permission(
        self, experiment_id: str, scorer_name: str, username: str, permission: str
    ) -> ScorerPermission:
        _validate_permission(permission)
        with self.ManagedSessionMaker() as session:
            perm = self._get_scorer_permission(session, experiment_id, scorer_name, username)
            perm.permission = permission
            return perm.to_mlflow_entity()

    def delete_scorer_permission(self, experiment_id: str, scorer_name: str, username: str):
        with self.ManagedSessionMaker() as session:
            perm = self._get_scorer_permission(session, experiment_id, scorer_name, username)
            session.delete(perm)

    def delete_scorer_permissions_for_scorer(self, experiment_id: str, scorer_name: str):
        with self.ManagedSessionMaker() as session:
            session.query(SqlScorerPermission).filter(
                SqlScorerPermission.experiment_id == experiment_id,
                SqlScorerPermission.scorer_name == scorer_name,
            ).delete()

    def create_gateway_secret_permission(
        self, secret_id: str, username: str, permission: str
    ) -> GatewaySecretPermission:
        _validate_permission(permission)
        with self.ManagedSessionMaker() as session:
            try:
                user = self._get_user(session, username=username)
                perm = SqlGatewaySecretPermission(
                    secret_id=secret_id, user_id=user.id, permission=permission
                )
                session.add(perm)
                session.flush()
                return perm.to_mlflow_entity()
            except IntegrityError as e:
                raise MlflowException(
                    f"Gateway secret permission (secret_id={secret_id}, username={username}) "
                    f"already exists. Error: {e}",
                    RESOURCE_ALREADY_EXISTS,
                ) from e

    def _get_gateway_secret_permission(
        self, session, secret_id: str, username: str
    ) -> SqlGatewaySecretPermission:
        try:
            user = self._get_user(session, username=username)
            return (
                session.query(SqlGatewaySecretPermission)
                .filter(
                    SqlGatewaySecretPermission.secret_id == secret_id,
                    SqlGatewaySecretPermission.user_id == user.id,
                )
                .one()
            )
        except NoResultFound:
            raise MlflowException(
                f"Gateway secret permission with secret_id={secret_id} and "
                f"username={username} not found",
                RESOURCE_DOES_NOT_EXIST,
            )
        except MultipleResultsFound:
            raise MlflowException(
                f"Found multiple gateway secret permissions with secret_id={secret_id} "
                f"and username={username}",
                INVALID_STATE,
            )

    def get_gateway_secret_permission(
        self, secret_id: str, username: str
    ) -> GatewaySecretPermission:
        with self.ManagedSessionMaker() as session:
            return self._get_gateway_secret_permission(
                session, secret_id, username
            ).to_mlflow_entity()

    def list_gateway_secret_permissions(self, username: str) -> list[GatewaySecretPermission]:
        with self.ManagedSessionMaker() as session:
            user = self._get_user(session, username=username)
            perms = (
                session.query(SqlGatewaySecretPermission)
                .filter(SqlGatewaySecretPermission.user_id == user.id)
                .all()
            )
            return [p.to_mlflow_entity() for p in perms]

    def update_gateway_secret_permission(
        self, secret_id: str, username: str, permission: str
    ) -> GatewaySecretPermission:
        _validate_permission(permission)
        with self.ManagedSessionMaker() as session:
            perm = self._get_gateway_secret_permission(session, secret_id, username)
            perm.permission = permission
            return perm.to_mlflow_entity()

    def delete_gateway_secret_permission(self, secret_id: str, username: str):
        with self.ManagedSessionMaker() as session:
            perm = self._get_gateway_secret_permission(session, secret_id, username)
            session.delete(perm)

    def delete_gateway_secret_permissions_for_secret(self, secret_id: str):
        with self.ManagedSessionMaker() as session:
            session.query(SqlGatewaySecretPermission).filter(
                SqlGatewaySecretPermission.secret_id == secret_id,
            ).delete()

    def create_gateway_endpoint_permission(
        self, endpoint_id: str, username: str, permission: str
    ) -> GatewayEndpointPermission:
        _validate_permission(permission)
        with self.ManagedSessionMaker() as session:
            try:
                user = self._get_user(session, username=username)
                perm = SqlGatewayEndpointPermission(
                    endpoint_id=endpoint_id, user_id=user.id, permission=permission
                )
                session.add(perm)
                session.flush()
                return perm.to_mlflow_entity()
            except IntegrityError as e:
                raise MlflowException(
                    f"Gateway endpoint permission (endpoint_id={endpoint_id}, username={username}) "
                    f"already exists. Error: {e}",
                    RESOURCE_ALREADY_EXISTS,
                ) from e

    def _get_gateway_endpoint_permission(
        self, session, endpoint_id: str, username: str
    ) -> SqlGatewayEndpointPermission:
        try:
            user = self._get_user(session, username=username)
            return (
                session.query(SqlGatewayEndpointPermission)
                .filter(
                    SqlGatewayEndpointPermission.endpoint_id == endpoint_id,
                    SqlGatewayEndpointPermission.user_id == user.id,
                )
                .one()
            )
        except NoResultFound:
            raise MlflowException(
                f"Gateway endpoint permission with endpoint_id={endpoint_id} and "
                f"username={username} not found",
                RESOURCE_DOES_NOT_EXIST,
            )
        except MultipleResultsFound:
            raise MlflowException(
                f"Found multiple gateway endpoint permissions with endpoint_id={endpoint_id} "
                f"and username={username}",
                INVALID_STATE,
            )

    def get_gateway_endpoint_permission(
        self, endpoint_id: str, username: str
    ) -> GatewayEndpointPermission:
        with self.ManagedSessionMaker() as session:
            return self._get_gateway_endpoint_permission(
                session, endpoint_id, username
            ).to_mlflow_entity()

    def list_gateway_endpoint_permissions(self, username: str) -> list[GatewayEndpointPermission]:
        with self.ManagedSessionMaker() as session:
            user = self._get_user(session, username=username)
            perms = (
                session.query(SqlGatewayEndpointPermission)
                .filter(SqlGatewayEndpointPermission.user_id == user.id)
                .all()
            )
            return [p.to_mlflow_entity() for p in perms]

    def update_gateway_endpoint_permission(
        self, endpoint_id: str, username: str, permission: str
    ) -> GatewayEndpointPermission:
        _validate_permission(permission)
        with self.ManagedSessionMaker() as session:
            perm = self._get_gateway_endpoint_permission(session, endpoint_id, username)
            perm.permission = permission
            return perm.to_mlflow_entity()

    def delete_gateway_endpoint_permission(self, endpoint_id: str, username: str):
        with self.ManagedSessionMaker() as session:
            perm = self._get_gateway_endpoint_permission(session, endpoint_id, username)
            session.delete(perm)

    def delete_gateway_endpoint_permissions_for_endpoint(self, endpoint_id: str):
        with self.ManagedSessionMaker() as session:
            session.query(SqlGatewayEndpointPermission).filter(
                SqlGatewayEndpointPermission.endpoint_id == endpoint_id,
            ).delete()

    def create_gateway_model_definition_permission(
        self, model_definition_id: str, username: str, permission: str
    ) -> GatewayModelDefinitionPermission:
        _validate_permission(permission)
        with self.ManagedSessionMaker() as session:
            try:
                user = self._get_user(session, username=username)
                perm = SqlGatewayModelDefinitionPermission(
                    model_definition_id=model_definition_id, user_id=user.id, permission=permission
                )
                session.add(perm)
                session.flush()
                return perm.to_mlflow_entity()
            except IntegrityError as e:
                raise MlflowException(
                    f"Gateway model definition permission "
                    f"(model_definition_id={model_definition_id}, username={username}) "
                    f"already exists. Error: {e}",
                    RESOURCE_ALREADY_EXISTS,
                ) from e

    def _get_gateway_model_definition_permission(
        self, session, model_definition_id: str, username: str
    ) -> SqlGatewayModelDefinitionPermission:
        try:
            user = self._get_user(session, username=username)
            return (
                session.query(SqlGatewayModelDefinitionPermission)
                .filter(
                    SqlGatewayModelDefinitionPermission.model_definition_id == model_definition_id,
                    SqlGatewayModelDefinitionPermission.user_id == user.id,
                )
                .one()
            )
        except NoResultFound:
            raise MlflowException(
                f"Gateway model definition permission with "
                f"model_definition_id={model_definition_id} and username={username} not found",
                RESOURCE_DOES_NOT_EXIST,
            )
        except MultipleResultsFound:
            raise MlflowException(
                f"Found multiple gateway model definition permissions with "
                f"model_definition_id={model_definition_id} and username={username}",
                INVALID_STATE,
            )

    def get_gateway_model_definition_permission(
        self, model_definition_id: str, username: str
    ) -> GatewayModelDefinitionPermission:
        with self.ManagedSessionMaker() as session:
            return self._get_gateway_model_definition_permission(
                session, model_definition_id, username
            ).to_mlflow_entity()

    def list_gateway_model_definition_permissions(
        self, username: str
    ) -> list[GatewayModelDefinitionPermission]:
        with self.ManagedSessionMaker() as session:
            user = self._get_user(session, username=username)
            perms = (
                session.query(SqlGatewayModelDefinitionPermission)
                .filter(SqlGatewayModelDefinitionPermission.user_id == user.id)
                .all()
            )
            return [p.to_mlflow_entity() for p in perms]

    def update_gateway_model_definition_permission(
        self, model_definition_id: str, username: str, permission: str
    ) -> GatewayModelDefinitionPermission:
        _validate_permission(permission)
        with self.ManagedSessionMaker() as session:
            perm = self._get_gateway_model_definition_permission(
                session, model_definition_id, username
            )
            perm.permission = permission
            return perm.to_mlflow_entity()

    def delete_gateway_model_definition_permission(self, model_definition_id: str, username: str):
        with self.ManagedSessionMaker() as session:
            perm = self._get_gateway_model_definition_permission(
                session, model_definition_id, username
            )
            session.delete(perm)

    def delete_gateway_model_definition_permissions_for_model_definition(
        self, model_definition_id: str
    ):
        with self.ManagedSessionMaker() as session:
            session.query(SqlGatewayModelDefinitionPermission).filter(
                SqlGatewayModelDefinitionPermission.model_definition_id == model_definition_id,
            ).delete()



import errno
import hashlib
import importlib
import inspect
import json
import logging
import os
import shutil
import signal
import subprocess
import sys
import tempfile
import threading
import time
from dataclasses import asdict, dataclass
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable

from mlflow.entities._job_status import JobStatus
from mlflow.environment_variables import (
    MLFLOW_SERVER_JOB_TRANSIENT_ERROR_RETRY_BASE_DELAY,
    MLFLOW_SERVER_JOB_TRANSIENT_ERROR_RETRY_MAX_DELAY,
)
from mlflow.exceptions import MlflowException
from mlflow.server.constants import HUEY_STORAGE_PATH_ENV_VAR
from mlflow.utils.environment import _PythonEnv
from mlflow.utils.import_hooks import register_post_import_hook
from mlflow.utils.process import _exec_cmd

if TYPE_CHECKING:
    import huey

    from mlflow.store.jobs.abstract_store import AbstractJobStore

_logger = logging.getLogger(__name__)

# Reserved Huey instance key for periodic tasks
HUEY_PERIODIC_TASKS_INSTANCE_KEY = "periodic_tasks"

# Number of worker threads for the periodic tasks consumer
PERIODIC_TASKS_WORKER_COUNT = 5


def _exponential_backoff_retry(retry_count: int) -> None:
    from huey.exceptions import RetryTask

    # We can support more retry strategies (e.g. exponential backoff) in future
    base_delay = MLFLOW_SERVER_JOB_TRANSIENT_ERROR_RETRY_BASE_DELAY.get()
    max_delay = MLFLOW_SERVER_JOB_TRANSIENT_ERROR_RETRY_MAX_DELAY.get()
    delay = min(base_delay * (2 ** (retry_count - 1)), max_delay)
    raise RetryTask(delay=delay)


@dataclass
class JobResult:
    succeeded: bool
    result: str | None = None  # serialized JSON string
    is_transient_error: bool | None = None
    error: str | None = None

    @classmethod
    def from_error(
        cls, e: Exception, transient_error_classes: list[type[Exception]] | None = None
    ) -> "JobResult":
        from mlflow.server.jobs import TransientError

        if isinstance(e, TransientError):
            return JobResult(succeeded=False, is_transient_error=True, error=repr(e.origin_error))

        if transient_error_classes:
            if e.__class__ in transient_error_classes:
                return JobResult(succeeded=False, is_transient_error=True, error=repr(e))

        return JobResult(
            succeeded=False,
            is_transient_error=False,
            error=repr(e),
        )

    def dump(self, path: str) -> None:
        with open(path, "w") as fp:
            json.dump(asdict(self), fp)

    @classmethod
    def load(cls, path: str) -> "JobResult":
        with open(path) as fp:
            return JobResult(**json.load(fp))


def _exit_when_orphaned(poll_interval: float = 1) -> None:
    while True:
        if os.getppid() == 1:
            os._exit(1)
        time.sleep(poll_interval)


def is_process_alive(pid: int) -> bool:
    if pid <= 0:
        return False
    try:
        os.kill(pid, 0)  # doesn't actually kill
    except OSError as e:
        if e.errno == errno.ESRCH:  # No such process
            return False
        elif e.errno == errno.EPERM:  # Process exists, but no permission
            return True
        else:
            raise
    else:
        return True


def _start_huey_consumer_proc(
    huey_instance_key: str,
    max_job_parallelism: int,
):
    from mlflow.server.constants import MLFLOW_HUEY_INSTANCE_KEY
    from mlflow.utils.process import _exec_cmd

    return _exec_cmd(
        [
            sys.executable,
            shutil.which("huey_consumer.py"),
            "mlflow.server.jobs._huey_consumer.huey_instance",
            "-w",
            str(max_job_parallelism),
        ],
        capture_output=False,
        synchronous=False,
        extra_env={
            MLFLOW_HUEY_INSTANCE_KEY: huey_instance_key,
        },
    )


_JOB_ENTRY_MODULE = "mlflow.server.jobs._job_subproc_entry"


_JOB_STATUS_POLL_INTERVAL = 1


def _exec_job_in_subproc(
    function_fullname: str,
    params: dict[str, Any],
    python_env: _PythonEnv | None,
    transient_error_classes: list[type[Exception]] | None,
    timeout: float | None,
    tmpdir: str,
    job_store: "AbstractJobStore",
    job_id: str,
) -> JobResult | None:
    """
    Executes the job function in a subprocess,
    If the job execution time exceeds timeout, the subprocess is killed and return None,
    otherwise return `JobResult` instance,
    """
    from mlflow.utils.process import _exec_cmd, _join_commands
    from mlflow.utils.virtualenv import (
        _get_mlflow_virtualenv_root,
        _get_uv_env_creation_command,
        _get_virtualenv_activate_cmd,
        _get_virtualenv_extra_env_vars,
        _get_virtualenv_name,
    )

    if python_env is not None:
        if shutil.which("uv") is None:
            raise MlflowException(
                "The job requires 'uv' to create an isolated Python environment, "
                "but 'uv' is not installed."
            )

        # set up virtual python environment
        virtual_envs_root_path = Path(_get_mlflow_virtualenv_root())
        env_name = _get_virtualenv_name(python_env, None)
        env_dir = virtual_envs_root_path / env_name
        activate_cmd = _get_virtualenv_activate_cmd(env_dir)

        if not env_dir.exists():
            _logger.info(f"Creating a python virtual environment in {env_dir}.")
            # create python environment
            env_creation_cmd = _get_uv_env_creation_command(env_dir, python_env.python)
            _exec_cmd(env_creation_cmd, capture_output=False)

            # install dependencies
            tmp_req_file = "requirements.txt"
            (Path(tmpdir) / tmp_req_file).write_text("\n".join(python_env.dependencies))
            cmd = _join_commands(activate_cmd, f"uv pip install -r {tmp_req_file}")
            _exec_cmd(
                cmd,
                cwd=tmpdir,
                extra_env=_get_virtualenv_extra_env_vars(),
                capture_output=False,
            )
        else:
            _logger.debug(f"The python environment {env_dir} already exists.")

        job_cmd = _join_commands(activate_cmd, f"exec python -m {_JOB_ENTRY_MODULE}")
    else:
        job_cmd = [sys.executable, "-m", _JOB_ENTRY_MODULE]

    result_file = str(Path(tmpdir) / "result.json")
    transient_error_classes_file = str(Path(tmpdir) / "transient_error_classes")
    transient_error_classes = transient_error_classes or []
    with open(transient_error_classes_file, "w") as f:
        for cls in transient_error_classes:
            f.write(f"{cls.__module__}.{cls.__name__}\n")

    with subprocess.Popen(
        job_cmd,
        env={
            **os.environ,
            "_MLFLOW_SERVER_JOB_PARAMS": json.dumps(params),
            "_MLFLOW_SERVER_JOB_FUNCTION_FULLNAME": function_fullname,
            "_MLFLOW_SERVER_JOB_RESULT_DUMP_PATH": result_file,
            "_MLFLOW_SERVER_JOB_TRANSIENT_ERROR_ClASSES_PATH": transient_error_classes_file,
        },
    ) as popen:
        beg_time = time.time()
        while popen.poll() is None:
            time.sleep(_JOB_STATUS_POLL_INTERVAL)

            job_status = job_store.get_job(job_id).status
            if job_status == JobStatus.CANCELED:
                popen.kill()
                return None

            if timeout is not None:
                if beg_time + timeout <= time.time():
                    # timeout
                    popen.kill()
                    job_store.mark_job_timed_out(job_id)
                    return None

        if popen.returncode == 0:
            return JobResult.load(result_file)

        return JobResult.from_error(
            RuntimeError(
                f"The subprocess that executes job function {function_fullname} "
                f"exists with error code {popen.returncode}"
            )
        )


def _compute_exclusive_lock_key(job_name: str, params: dict[str, Any]) -> str:
    """
    Compute a lock key based on job name and params hash.

    Args:
        job_name: Name of the job.
        params: Parameter dictionary to use for the lock key.

    Returns:
        Lock key string.
    """
    params_json = json.dumps(params, sort_keys=True)
    params_hash = hashlib.sha256(params_json.encode()).hexdigest()[:16]
    return f"{job_name}:{params_hash}"


def _exec_job(
    job_id: str,
    job_name: str,
    params: dict[str, Any],
    timeout: float | None,
    exclusive: bool | list[str] = False,
) -> None:
    """
    Execute a job in a subprocess.

    Args:
        job_id: Unique identifier for the job.
        job_name: Name of the job function to execute.
        params: Parameters to pass to the job function.
        timeout: Maximum execution time in seconds, or None for no timeout.
        exclusive: If True, only one instance of this job with the same params can run
            at a time. If a list of parameter names, only those parameters are considered
            for exclusivity.
    """
    from mlflow.server.handlers import _get_job_store

    job_store = _get_job_store()

    # If exclusive, acquire lock based on job_name + hash(params)
    # If lock is already held, TaskLockedException is raised and job is skipped
    if exclusive:
        from huey.exceptions import TaskLockedException

        huey_instance = _get_or_init_huey_instance(job_name).instance
        # If exclusive is a list, filter params to only those specified
        lock_params = (
            {k: v for k, v in params.items() if k in exclusive}
            if isinstance(exclusive, list)
            else params
        )
        lock_key = _compute_exclusive_lock_key(job_name, lock_params)
        lock = huey_instance.lock_task(lock_key)
        try:
            lock.acquire()
        except TaskLockedException:
            _logger.info(f"Skipping job {job_id} - exclusive lock {lock_key} already held")
            job_store.cancel_job(job_id)
            return
    else:
        lock = None

    try:
        job_store.start_job(job_id)

        fn_fullname = get_job_fn_fullname(job_name)
        function = _load_function(fn_fullname)
        fn_metadata = function._job_fn_metadata

        with tempfile.TemporaryDirectory() as tmpdir:
            job_result = _exec_job_in_subproc(
                fn_metadata.fn_fullname,
                params,
                fn_metadata.python_env,
                fn_metadata.transient_error_classes,
                timeout,
                tmpdir,
                job_store,
                job_id,
            )

        if job_result is None:
            return

        if job_result.succeeded:
            job_store.finish_job(job_id, job_result.result)
            return

        if job_result.is_transient_error:
            # For transient errors, if the retry count is less than max allowed count,
            # trigger task retry by raising `RetryTask` exception.
            retry_count = job_store.retry_or_fail_job(job_id, job_result.error)
            if retry_count is not None:
                _exponential_backoff_retry(retry_count)
        else:
            job_store.fail_job(job_id, job_result.error)
    finally:
        if lock is not None:
            lock.release()


@dataclass
class HueyInstance:
    instance: "huey.SqliteHuey"
    submit_task: Callable[..., Any]


# Each job function has an individual execution pool, each execution pool
# is managed by a Huey instance.
# The `_huey_instance_map` stores the map, the key is the job function fullname,
# and the value is the `HueyInstance` object.
_huey_instance_map: dict[str, HueyInstance] = {}
_huey_instance_map_lock = threading.RLock()


def _get_or_init_huey_instance(instance_key: str):
    from huey import SqliteHuey
    from huey.serializer import Serializer

    class CustomJSONEncoder(json.JSONEncoder):
        def default(self, obj):
            if isinstance(obj, datetime):
                return {
                    "__type__": "datetime",
                    "value": obj.isoformat(),
                }
            return super().default(obj)

    def json_loader_object_hook(d):
        if d.get("__type__") == "datetime":
            return datetime.fromisoformat(d["value"])
        return d

    class JsonSerializer(Serializer):
        def serialize(self, data):
            # Huey passes two types of data through the serializer:
            # 1. Message objects (task data) - have ._asdict() method
            # 2. Plain data (e.g., lock values like '1') - no ._asdict() method
            # We need to handle both cases for exclusive job locks to work.
            data_dict = data._asdict() if hasattr(data, "_asdict") else data
            return json.dumps(data_dict, cls=CustomJSONEncoder).encode("utf-8")

        def deserialize(self, data):
            from huey.registry import Message

            decoded = json.loads(data.decode("utf-8"), object_hook=json_loader_object_hook)
            # Message objects have specific structure: {"id": ..., "name": ..., ...}
            # Only reconstruct as Message when that structure exists.
            # Plain data (like lock values) should be returned as-is.
            if isinstance(decoded, dict) and "id" in decoded and "name" in decoded:
                return Message(**decoded)
            else:
                return decoded

    with _huey_instance_map_lock:
        if instance_key not in _huey_instance_map:
            _logger.info(f"Creating huey instance for {instance_key}")
            huey_store_file = os.path.join(
                os.environ[HUEY_STORAGE_PATH_ENV_VAR], f"{instance_key}.mlflow-huey-store"
            )
            huey_instance = SqliteHuey(
                filename=huey_store_file,
                results=False,
                serializer=JsonSerializer(),
            )
            huey_submit_task_fn = huey_instance.task(retries=0)(_exec_job)
            _huey_instance_map[instance_key] = HueyInstance(
                instance=huey_instance,
                submit_task=huey_submit_task_fn,
            )
        return _huey_instance_map[instance_key]


def _launch_huey_consumer(job_name: str) -> None:
    _logger.info(f"Starting huey consumer for job function {job_name}")

    fn_fullname = get_job_fn_fullname(job_name)
    job_fn = _load_function(fn_fullname)

    if not hasattr(job_fn, "_job_fn_metadata"):
        raise MlflowException.invalid_parameter_value(
            f"The job function {job_name} is not decorated by 'mlflow.server.jobs.job_function'."
        )

    max_job_parallelism = job_fn._job_fn_metadata.max_workers

    def _huey_consumer_thread() -> None:
        while True:
            # start MLflow job runner process
            # Put it inside the loop to ensure the job runner process alive
            job_runner_proc = _start_huey_consumer_proc(
                job_name,
                max_job_parallelism,
            )
            job_runner_proc.wait()
            time.sleep(1)

    # start job runner.
    threading.Thread(
        target=_huey_consumer_thread,
        name=f"MLflow-huey-consumer-{job_name}-watcher",
        daemon=False,
    ).start()


def _launch_periodic_tasks_consumer() -> None:
    """
    Launch a dedicated Huey consumer for periodic tasks.
    This consumer runs scheduled tasks like the online scoring scheduler.
    """
    _logger.info("Starting dedicated Huey consumer for periodic tasks")

    def _huey_consumer_thread() -> None:
        while True:
            job_runner_proc = _start_periodic_tasks_consumer_proc()
            job_runner_proc.wait()
            time.sleep(1)

    threading.Thread(
        target=_huey_consumer_thread,
        name="MLflow-huey-consumer-periodic-tasks-watcher",
        daemon=False,
    ).start()


def _start_periodic_tasks_consumer_proc():
    return _exec_cmd(
        [
            sys.executable,
            shutil.which("huey_consumer.py"),
            "mlflow.server.jobs._periodic_tasks_consumer.huey_instance",
            "-w",
            str(PERIODIC_TASKS_WORKER_COUNT),
        ],
        capture_output=False,
        synchronous=False,
    )


def _launch_job_runner(env_map, server_proc_pid):
    return subprocess.Popen(
        [
            sys.executable,
            "-m",
            "mlflow.server.jobs._job_runner",
        ],
        env={**os.environ, **env_map, "MLFLOW_SERVER_PID": str(server_proc_pid)},
    )


def _start_watcher_to_kill_job_runner_if_mlflow_server_dies(check_interval: float = 1.0) -> None:
    mlflow_server_pid = int(os.environ.get("MLFLOW_SERVER_PID"))

    def watcher():
        while True:
            if not is_process_alive(mlflow_server_pid):
                os.kill(os.getpid(), signal.SIGTERM)
            time.sleep(check_interval)

    t = threading.Thread(target=watcher, daemon=True, name="job-runner-watcher")
    t.start()


def _load_function(fullname: str) -> Callable[..., Any]:
    match fullname.split("."):
        case [*module_parts, func_name] if module_parts:
            module_name = ".".join(module_parts)
        case _:
            raise MlflowException.invalid_parameter_value(
                f"Invalid function fullname format: {fullname}"
            )
    try:
        module = importlib.import_module(module_name)
        return getattr(module, func_name)
    except ModuleNotFoundError:
        # Module doesn't exist
        raise MlflowException.invalid_parameter_value(
            f"Module not found for function '{fullname}'",
        )
    except AttributeError:
        # Function doesn't exist in the module
        raise MlflowException.invalid_parameter_value(
            f"Function not found in module for '{fullname}'",
        )


def _enqueue_unfinished_jobs(server_launching_timestamp: int) -> None:
    from mlflow.server.handlers import _get_job_store

    job_store = _get_job_store()

    unfinished_jobs = job_store.list_jobs(
        statuses=[JobStatus.PENDING, JobStatus.RUNNING],
        # filter out jobs created after the server is launched.
        end_timestamp=server_launching_timestamp,
    )

    for job in unfinished_jobs:
        if job.status == JobStatus.RUNNING:
            job_store.reset_job(job.job_id)  # reset the job status to PENDING

        params = json.loads(job.params)
        timeout = job.timeout
        # Look up exclusive flag from function metadata
        fn_fullname = get_job_fn_fullname(job.job_name)
        fn_metadata = _load_function(fn_fullname)._job_fn_metadata
        _get_or_init_huey_instance(job.job_name).submit_task(
            job.job_id, job.job_name, params, timeout, fn_metadata.exclusive
        )


def _validate_function_parameters(function: Callable[..., Any], params: dict[str, Any]) -> None:
    """Validate that the provided parameters match the function's required arguments.

    Args:
        function: The function to validate parameters against
        params: Dictionary of parameters provided for the function

    Raises:
        MlflowException: If required parameters are missing
    """
    sig = inspect.signature(function)

    # Get all required parameters (no default value)
    # Exclude VAR_POSITIONAL (*args) and VAR_KEYWORD (**kwargs) parameters
    required_params = [
        name
        for name, param in sig.parameters.items()
        if param.default is inspect.Parameter.empty
        and param.kind not in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD)
    ]

    # Check for missing required parameters
    if missing_params := [param for param in required_params if param not in params]:
        raise MlflowException.invalid_parameter_value(
            f"Missing required parameters for function '{function.__name__}': {missing_params}. "
            f"Expected parameters: {list(sig.parameters.keys())}"
        )


def _check_requirements(backend_store_uri: str | None = None) -> None:
    from mlflow.server.constants import BACKEND_STORE_URI_ENV_VAR
    from mlflow.utils.uri import extract_db_type_from_uri

    if os.name == "nt":
        raise MlflowException("MLflow job backend does not support Windows system.")

    backend_store_uri = backend_store_uri or os.environ.get(BACKEND_STORE_URI_ENV_VAR)
    if not backend_store_uri:
        raise MlflowException(
            "MLflow job backend requires a database backend store URI but "
            "'--backend-store-uri' is not set"
        )
    try:
        extract_db_type_from_uri(backend_store_uri)
    except MlflowException:
        raise MlflowException(
            f"MLflow job backend requires a database backend store URI but got {backend_store_uri}"
        )


# The map from job name to the job function's fullname.
_job_name_to_fn_fullname_map = {}


def get_job_fn_fullname(job_name: str):
    if job_name not in _job_name_to_fn_fullname_map:
        raise MlflowException.invalid_parameter_value(f"Invalid job name: {job_name}")
    return _job_name_to_fn_fullname_map[job_name]


def _build_job_name_to_fn_fullname_map():
    from mlflow.server.jobs import _SUPPORTED_JOB_FUNCTION_LIST

    for fn_fullname in set(_SUPPORTED_JOB_FUNCTION_LIST):
        try:
            fn_meta = _load_function(fn_fullname)._job_fn_metadata
            if exist_fullname := _job_name_to_fn_fullname_map.get(fn_meta.name):
                if exist_fullname != fn_fullname:
                    _logger.warning(
                        f"The 2 job functions {fn_fullname} and {exist_fullname} have the same "
                        f"job name {fn_meta.name}, this is not allowed, skip loading function "
                        f"{fn_fullname}."
                    )
            else:
                _job_name_to_fn_fullname_map[fn_meta.name] = fn_fullname
        except Exception as e:
            _logger.warning(f"loading job function {fn_fullname} failed: {e!r}", exc_info=True)


register_post_import_hook(lambda m: _build_job_name_to_fn_fullname_map(), __name__)


def register_periodic_tasks(huey_instance) -> None:
    """
    Register all periodic tasks with the given huey instance.

    Args:
        huey_instance: The huey instance to register tasks with.
    """
    from huey import crontab

    @huey_instance.periodic_task(crontab(minute="*/1"))
    # Prevent concurrent execution if scheduler takes longer than 1 minute.
    @huey_instance.lock_task("online-scoring-scheduler-lock")
    def online_scoring_scheduler():
        """Runs every minute to fetch active scorer configs and submit scoring jobs."""
        from mlflow.genai.scorers.job import run_online_scoring_scheduler

        try:
            run_online_scoring_scheduler()
        except Exception as e:
            _logger.exception(f"Online scoring scheduler failed: {e!r}")

    _logger.info("Registered online_scoring_scheduler periodic task (runs every 1 minute)")



import json
import os
from collections.abc import Generator
from functools import lru_cache
from pathlib import Path
from typing import Any
from urllib.request import urlopen

from mlflow.entities import SourceType
from mlflow.tracking.context.abstract_context import RunContextProvider
from mlflow.utils.databricks_utils import is_running_in_ipython_environment
from mlflow.utils.mlflow_tags import MLFLOW_SOURCE_NAME, MLFLOW_SOURCE_TYPE


@lru_cache(maxsize=1)
def _get_notebook_name() -> str | None:
    """
    Attempt to get the current Jupyter notebook name using multiple methods.

    Returns:
        The notebook filename if found, None otherwise.
    """
    # Method 1: Check VS Code notebook path in IPython user namespace
    # VS Code's Jupyter extension stores the notebook path in ip.user_ns
    if notebook_path := _get_vscode_notebook_path():
        return os.path.basename(notebook_path)

    # Method 2: Check environment variables
    if vsc_notebook := os.environ.get("__vsc_ipynb_file__"):
        return os.path.basename(vsc_notebook)

    if ipynb_file := os.environ.get("IPYNB_FILE"):
        return os.path.basename(ipynb_file)

    # Method 3: Query Jupyter server sessions (for JupyterLab/classic Jupyter)
    if notebook_path := _get_notebook_path_from_sessions():
        return os.path.basename(notebook_path)

    return None


def _get_vscode_notebook_path() -> str | None:
    """
    Get notebook path from VS Code's IPython user namespace.

    VS Code's Jupyter extension stores the notebook file path in the
    IPython user namespace under '__vsc_ipynb_file__'.

    Returns:
        The notebook path if found, None otherwise.
    """
    if not is_running_in_ipython_environment():
        return None

    try:
        from IPython import get_ipython

        ip = get_ipython()
        if ip and "__vsc_ipynb_file__" in ip.user_ns:
            return ip.user_ns["__vsc_ipynb_file__"]
    except Exception:
        pass

    return None


def _get_notebook_path_from_sessions() -> str | None:
    """
    Get notebook path by querying Jupyter server sessions.

    This queries running Jupyter servers to find the notebook
    associated with the current kernel.

    Returns:
        The notebook path if found, None otherwise.
    """
    try:
        kernel_id = _get_kernel_id()
        if not kernel_id:
            return None

        for server in _get_running_servers():
            try:
                if notebook_path := _get_sessions_notebook(server, kernel_id):
                    return notebook_path
            except Exception:
                continue
    except Exception:
        pass

    return None


def _get_kernel_id() -> str | None:
    """
    Get the current kernel ID from the connection file.

    Returns:
        The kernel ID string, or None if not found.
    """
    try:
        import ipykernel

        connection_file = Path(ipykernel.get_connection_file()).stem
        # Connection file is like: kernel-<uuid>
        return connection_file.split("-", 1)[1]
    except Exception:
        pass

    return None


def _get_running_servers() -> Generator[dict[str, Any], None, None]:
    """
    Get list of running Jupyter servers by scanning the runtime directory.

    Yields:
        Server info dictionaries with 'url' and 'token' keys.
    """
    try:
        from jupyter_core.paths import jupyter_runtime_dir

        runtime_dir = Path(jupyter_runtime_dir())
        if not runtime_dir.is_dir():
            return

        # Get server files, sorted by modification time (most recent first)
        server_files = sorted(
            list(runtime_dir.glob("nbserver-*.json"))  # jupyter notebook (or lab 2)
            + list(runtime_dir.glob("jpserver-*.json")),  # jupyterlab 3
            key=os.path.getmtime,
            reverse=True,
        )

        for server_file in server_files:
            try:
                with open(server_file) as f:
                    yield json.load(f)
            except (json.JSONDecodeError, OSError):
                continue
    except ImportError:
        pass


def _get_sessions_notebook(server: dict[str, Any], kernel_id: str) -> str | None:
    """
    Query a server's sessions API to find the notebook for a kernel.

    Args:
        server: Server info dict with 'url' and optionally 'token'.
        kernel_id: The kernel ID to search for.

    Returns:
        The notebook path if found, None otherwise.
    """
    url = server.get("url", "").rstrip("/")
    token = server.get("token") or os.getenv("JUPYTERHUB_API_TOKEN", "")

    sessions_url = f"{url}/api/sessions"
    if token:
        sessions_url += f"?token={token}"

    try:
        with urlopen(sessions_url, timeout=0.5) as response:
            sessions = json.load(response)

        for session in sessions:
            if session.get("kernel", {}).get("id") == kernel_id:
                return session.get("path")
    except Exception:
        pass

    return None


def _is_in_jupyter_notebook() -> bool:
    """
    Check if we're running inside a Jupyter notebook (not just IPython).

    Returns:
        True if we're in a Jupyter notebook environment, False otherwise.
    """
    if not is_running_in_ipython_environment():
        return False

    try:
        from IPython import get_ipython

        ip = get_ipython()

        # Jupyter notebooks use ZMQInteractiveShell
        shell_class = ip.__class__.__name__
        if shell_class == "ZMQInteractiveShell":
            return True

        # Also check for kernel attribute
        if hasattr(ip, "kernel"):
            return True

        return False
    except (ImportError, ModuleNotFoundError):
        return False


class JupyterNotebookRunContext(RunContextProvider):
    """
    Context provider for local Jupyter notebooks.

    This provider sets the source name to the notebook filename and source type
    to NOTEBOOK when running inside a Jupyter notebook environment.
    """

    def in_context(self):
        return _is_in_jupyter_notebook()

    def tags(self):
        notebook_name = _get_notebook_name()
        tags = {
            MLFLOW_SOURCE_TYPE: SourceType.to_string(SourceType.NOTEBOOK),
        }
        if notebook_name:
            tags[MLFLOW_SOURCE_NAME] = notebook_name
        return tags



from abc import ABC, abstractmethod
from functools import lru_cache
from pathlib import Path
from typing import Any, AsyncGenerator, Callable

from mlflow.assistant.config import AssistantConfig, ProviderConfig


@lru_cache(maxsize=10)
def load_config(name: str) -> ProviderConfig:
    cfg = AssistantConfig.load()
    if not cfg or name not in cfg.providers:
        raise RuntimeError(f"Provider configuration not found for {name}")
    return cfg.providers[name]


class ProviderNotConfiguredError(Exception):
    """Raised when a provider is not properly configured."""


class CLINotInstalledError(ProviderNotConfiguredError):
    """Raised when the provider CLI is not installed."""


class NotAuthenticatedError(ProviderNotConfiguredError):
    """Raised when the user is not authenticated with the provider."""


class AssistantProvider(ABC):
    """Abstract base class for assistant providers."""

    @property
    @abstractmethod
    def name(self) -> str:
        """Return the provider identifier (e.g., 'claude_code')."""

    @property
    @abstractmethod
    def display_name(self) -> str:
        """Return the human-readable provider name (e.g., 'Claude Code')."""

    @property
    @abstractmethod
    def description(self) -> str:
        """Return a short description of the provider."""

    @abstractmethod
    def is_available(self) -> bool:
        """Check if the provider is available and ready to use."""

    @abstractmethod
    def check_connection(self, echo: Callable[[str], None] | None = None) -> None:
        """
        Check if the provider is properly configured and can connect.

        Args:
            echo: Optional function to print status messages.

        Raises:
            ProviderNotConfiguredError: If the provider is not properly configured.
        """

    @abstractmethod
    def install_skills(self, skill_path: Path) -> list[str]:
        """Install provider-specific skills.

        Args:
            skill_path: Directory where skills should be installed.

        Returns:
            List of installed skill names.
        """

    @abstractmethod
    def astream(
        self,
        prompt: str,
        session_id: str | None = None,
    ) -> AsyncGenerator[dict[str, Any], None]:
        """
        Stream responses from the assistant asynchronously.

        Args:
            prompt: The prompt to send to the assistant
            session_id: Session ID for conversation continuity

        Yields:
            Event dictionaries with 'type' and 'data' keys.
            Event types: 'message', 'status', 'done', 'error'
        """



import mlflow.pyfunc.loaders.chat_agent  # noqa: F401
import mlflow.pyfunc.loaders.chat_model  # noqa: F401
import mlflow.pyfunc.loaders.code_model  # noqa: F401
import mlflow.pyfunc.loaders.responses_agent  # noqa: F401



"""add_gateway_permissions

Revision ID: a1b2c3d4e5f6
Revises: 0965eb92f5f0
Create Date: 2025-12-30 12:00:00.000000

"""

import sqlalchemy as sa
from alembic import op

revision = "a1b2c3d4e5f6"
down_revision = "0965eb92f5f0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "gateway_secret_permissions",
        sa.Column("id", sa.Integer(), nullable=False, primary_key=True),
        sa.Column("secret_id", sa.String(length=255), nullable=False),
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column("permission", sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(["user_id"], ["users.id"], name="fk_gateway_secret_perm_user_id"),
        sa.UniqueConstraint("secret_id", "user_id", name="unique_secret_user"),
    )

    op.create_table(
        "gateway_endpoint_permissions",
        sa.Column("id", sa.Integer(), nullable=False, primary_key=True),
        sa.Column("endpoint_id", sa.String(length=255), nullable=False),
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column("permission", sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(["user_id"], ["users.id"], name="fk_gateway_endpoint_perm_user_id"),
        sa.UniqueConstraint("endpoint_id", "user_id", name="unique_endpoint_user"),
    )

    op.create_table(
        "gateway_model_definition_permissions",
        sa.Column("id", sa.Integer(), nullable=False, primary_key=True),
        sa.Column("model_definition_id", sa.String(length=255), nullable=False),
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column("permission", sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(
            ["user_id"], ["users.id"], name="fk_gateway_model_def_perm_user_id"
        ),
        sa.UniqueConstraint("model_definition_id", "user_id", name="unique_model_def_user"),
    )


def downgrade() -> None:
    op.drop_table("gateway_model_definition_permissions")
    op.drop_table("gateway_endpoint_permissions")
    op.drop_table("gateway_secret_permissions")



"""Huey job functions for async scorer invocation."""

import logging
import random
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import asdict, dataclass, field
from typing import Any

from mlflow.entities import Trace
from mlflow.environment_variables import (
    MLFLOW_GENAI_EVAL_MAX_WORKERS,
    MLFLOW_SERVER_JUDGE_INVOKE_MAX_WORKERS,
    MLFLOW_SERVER_ONLINE_SCORING_MAX_WORKERS,
    MLFLOW_SERVER_SCORER_INVOKE_BATCH_SIZE,
)
from mlflow.exceptions import MlflowException
from mlflow.genai.evaluation.entities import EvalItem
from mlflow.genai.evaluation.harness import _compute_eval_scores, _log_assessments
from mlflow.genai.evaluation.session_utils import (
    evaluate_session_level_scorers,
    get_first_trace_in_session,
)
from mlflow.genai.scorers.base import Scorer
from mlflow.genai.scorers.online import (
    OnlineScorer,
    OnlineScoringConfig,
    OnlineSessionScoringProcessor,
    OnlineTraceScoringProcessor,
)
from mlflow.server.handlers import _get_tracking_store
from mlflow.server.jobs import job, submit_job
from mlflow.store.tracking.abstract_store import AbstractStore
from mlflow.tracing.constant import TraceMetadataKey

_logger = logging.getLogger(__name__)


@dataclass
class ScorerFailure:
    error_code: str
    error_message: str


@dataclass
class TraceResult:
    assessments: list[Any] = field(default_factory=list)
    failures: list[ScorerFailure] = field(default_factory=list)


def _extract_failures_from_feedbacks(feedbacks: list[Any]) -> list[ScorerFailure]:
    return [
        ScorerFailure(
            error_code=feedback.error.error_code,
            error_message=feedback.error.error_message,
        )
        for feedback in feedbacks
        if feedback.error
    ]


@job(
    name="run_online_trace_scorer",
    max_workers=MLFLOW_SERVER_ONLINE_SCORING_MAX_WORKERS.get(),
    exclusive=["experiment_id"],
)
def run_online_trace_scorer_job(
    experiment_id: str,
    online_scorers: list[dict[str, Any]],
) -> None:
    """
    Job that fetches samples of individual traces and runs scorers on them.

    This job is exclusive per experiment_id to prevent duplicate scoring of the same
    experiment. Multiple jobs with different scorers for the same experiment will not
    run simultaneously, ensuring consistent checkpoint management.

    Args:
        experiment_id: The experiment ID to fetch traces from.
        online_scorers: List of OnlineScorer dicts specifying which scorers to run.
    """
    scorer_objects = [
        OnlineScorer(
            name=scorer_dict["name"],
            serialized_scorer=scorer_dict["serialized_scorer"],
            online_config=OnlineScoringConfig(**scorer_dict["online_config"]),
        )
        for scorer_dict in online_scorers
    ]

    tracking_store = _get_tracking_store()
    processor = OnlineTraceScoringProcessor.create(experiment_id, scorer_objects, tracking_store)
    processor.process_traces()


@job(
    name="run_online_session_scorer",
    max_workers=MLFLOW_SERVER_ONLINE_SCORING_MAX_WORKERS.get(),
    exclusive=["experiment_id"],
)
def run_online_session_scorer_job(
    experiment_id: str,
    online_scorers: list[dict[str, Any]],
) -> None:
    """
    Job that finds completed sessions and runs session-level scorers on them.

    This job is exclusive per experiment_id to prevent duplicate scoring of the same
    experiment. Multiple jobs with different scorers for the same experiment will not
    run simultaneously, ensuring consistent checkpoint management.

    Args:
        experiment_id: The experiment ID to fetch sessions from.
        online_scorers: List of OnlineScorer dicts specifying which scorers to run.
    """
    scorer_objects = [
        OnlineScorer(
            name=scorer_dict["name"],
            serialized_scorer=scorer_dict["serialized_scorer"],
            online_config=OnlineScoringConfig(**scorer_dict["online_config"]),
        )
        for scorer_dict in online_scorers
    ]

    tracking_store = _get_tracking_store()
    processor = OnlineSessionScoringProcessor.create(experiment_id, scorer_objects, tracking_store)
    processor.process_sessions()


@job(name="invoke_scorer", max_workers=MLFLOW_SERVER_JUDGE_INVOKE_MAX_WORKERS.get())
def invoke_scorer_job(
    experiment_id: str,
    serialized_scorer: str,
    trace_ids: list[str],
    log_assessments: bool = True,
) -> dict[str, Any]:
    """
    Huey job function for async scorer invocation.

    Reuses the core scoring logic from the evaluation harness
    for consistency and feature parity.

    Args:
        experiment_id: The experiment ID for the traces.
        serialized_scorer: JSON string of the serialized scorer.
        trace_ids: List of trace IDs to evaluate.
        log_assessments: Whether to log assessments to the traces.

    Returns:
        Dict mapping trace_id to TraceResult (assessments and failures).
    """
    # Deserialize scorer
    scorer = Scorer.model_validate_json(serialized_scorer)

    tracking_store = _get_tracking_store()

    if scorer.is_session_level_scorer:
        result = _run_session_scorer(scorer, trace_ids, tracking_store, log_assessments)
    else:
        result = _run_single_turn_scorer_batch(scorer, trace_ids, tracking_store, log_assessments)

    return {trace_id: asdict(trace_result) for trace_id, trace_result in result.items()}


def _fetch_traces_batch(
    trace_ids: list[str],
    tracking_store: AbstractStore,
) -> dict[str, Trace]:
    """
    Fetch traces in batch and return a mapping.

    Args:
        trace_ids: List of trace IDs to fetch.
        tracking_store: The tracking store instance.

    Returns:
        Dict mapping trace_id to Trace.

    Raises:
        MlflowException: If any trace IDs are not found.
    """
    traces = tracking_store.batch_get_traces(trace_ids)
    trace_map = {t.info.trace_id: t for t in traces}

    if missing_ids := [tid for tid in trace_ids if tid not in trace_map]:
        raise MlflowException(f"Traces not found: {missing_ids}")

    return trace_map


def _run_session_scorer(
    scorer: Any,
    trace_ids: list[str],
    tracking_store: AbstractStore,
    log_assessments: bool,
) -> dict[str, TraceResult]:
    """
    Run a session-level scorer on all traces as a conversation.

    Reuses evaluate_session_level_scorers from the evaluation harness.

    Args:
        scorer: The scorer instance.
        trace_ids: List of trace IDs (representing a session).
        tracking_store: The tracking store instance.
        log_assessments: Whether to log assessments to the traces.

    Returns:
        Dict mapping trace_id to TraceResult.
    """
    trace_map = _fetch_traces_batch(trace_ids, tracking_store)

    # Preserve order of traces as requested
    traces = [trace_map[tid] for tid in trace_ids]

    session_items = [EvalItem.from_trace(t) for t in traces]

    # Get session_id from the first trace's metadata
    first_session_item = get_first_trace_in_session(session_items)
    trace_metadata = first_session_item.trace.info.trace_metadata or {}
    session_id = trace_metadata.get(TraceMetadataKey.TRACE_SESSION)

    if not session_id:
        raise MlflowException(
            "Session-level scorer requires traces with session metadata. "
            f"Trace {first_session_item.trace.info.trace_id} is missing "
            f"'{TraceMetadataKey.TRACE_SESSION}' in its metadata."
        )

    first_trace = first_session_item.trace
    first_trace_id = first_trace.info.trace_id

    try:
        result = evaluate_session_level_scorers(
            session_id=session_id,
            session_items=session_items,
            multi_turn_scorers=[scorer],
        )

        # result is {first_trace_id: [feedbacks]}
        feedbacks = result[first_trace_id]

        failures = _extract_failures_from_feedbacks(feedbacks)

        if log_assessments and feedbacks:
            _log_assessments(
                run_id=None,  # No MLflow run context in API path
                trace=first_trace,
                assessments=feedbacks,
            )

        return {
            first_trace_id: TraceResult(
                assessments=[f.to_dictionary() for f in feedbacks],
                failures=failures,
            )
        }
    except Exception as e:
        return {
            first_trace_id: TraceResult(
                failures=[ScorerFailure(error_code=type(e).__name__, error_message=str(e))]
            )
        }


def _run_single_turn_scorer_batch(
    scorer: Any,
    trace_ids: list[str],
    tracking_store: AbstractStore,
    log_assessments: bool,
) -> dict[str, TraceResult]:
    """
    Run a single-turn scorer on each trace in parallel.

    Args:
        scorer: The scorer instance.
        trace_ids: List of trace IDs to evaluate.
        tracking_store: The tracking store instance.
        log_assessments: Whether to log assessments to the traces.

    Returns:
        Dict mapping trace_id to TraceResult.
    """
    trace_map = _fetch_traces_batch(trace_ids, tracking_store)

    def process_trace(trace_id: str, trace: Trace) -> tuple[str, TraceResult]:
        eval_item = EvalItem.from_trace(trace)

        try:
            # Use _compute_eval_scores from harness - supports scorer tracing,
            # captures stack traces on errors
            feedbacks = _compute_eval_scores(
                eval_item=eval_item,
                scorers=[scorer],
            )

            failures = _extract_failures_from_feedbacks(feedbacks)

            if log_assessments and feedbacks:
                _log_assessments(
                    run_id=None,  # No MLflow run context in API path
                    trace=trace,
                    assessments=feedbacks,
                )

            return trace_id, TraceResult(
                assessments=[f.to_dictionary() for f in feedbacks],
                failures=failures,
            )
        except Exception as e:
            return trace_id, TraceResult(
                failures=[ScorerFailure(error_code=type(e).__name__, error_message=str(e))]
            )

    max_workers = min(len(trace_map), MLFLOW_GENAI_EVAL_MAX_WORKERS.get())
    results: dict[str, TraceResult] = {}

    with ThreadPoolExecutor(
        max_workers=max_workers,
        thread_name_prefix="MlflowScorerInvoke",
    ) as executor:
        futures = {
            executor.submit(process_trace, tid, trace): tid for tid, trace in trace_map.items()
        }
        for future in as_completed(futures):
            trace_id, result = future.result()
            results[trace_id] = result

    return results


def _group_traces_by_session_id(
    trace_ids: list[str],
    tracking_store: AbstractStore,
) -> dict[str, list[str]]:
    """
    Group trace_ids by their session_id metadata.

    Fetches trace info from the tracking store and groups them by session_id.
    Traces without a session_id are skipped.

    Args:
        trace_ids: List of trace IDs to group.
        tracking_store: The tracking store instance.

    Returns:
        Dictionary mapping session_id to list of trace_ids, sorted by timestamp.
    """
    session_groups: dict[str, list[str]] = {}
    # trace_id -> (session_id, timestamp_ms)
    trace_info_cache: dict[str, tuple[str, int | None]] = {}

    trace_infos = tracking_store.batch_get_trace_infos(trace_ids)

    for trace_info in trace_infos:
        trace_metadata = trace_info.trace_metadata or {}
        if session_id := trace_metadata.get(TraceMetadataKey.TRACE_SESSION):
            if session_id not in session_groups:
                session_groups[session_id] = []
            session_groups[session_id].append(trace_info.trace_id)
            trace_info_cache[trace_info.trace_id] = (session_id, trace_info.timestamp_ms)

    # Sort trace_ids within each session by trace timestamp (None timestamps sort last)
    for session_id in session_groups:
        session_groups[session_id] = sorted(
            session_groups[session_id],
            key=lambda tid: trace_info_cache.get(tid, ("", None))[1] or float("inf"),
        )

    return session_groups


def get_trace_batches_for_scorer(
    trace_ids: list[str],
    scorer: Scorer,
    tracking_store: AbstractStore,
) -> list[list[str]]:
    """
    Get trace ID batches for scorer invocation.

    Handles batching logic for both session-level and single-turn scorers:
    - Session-level scorers: Groups traces by session_id, returns one batch per session.
    - Single-turn scorers: Batches traces based on MLFLOW_SERVER_SCORER_INVOKE_BATCH_SIZE.

    Args:
        trace_ids: List of trace IDs to evaluate.
        scorer: The validated Scorer instance.
        tracking_store: The tracking store instance.

    Returns:
        List of trace ID batches, where each batch should be submitted as a separate job.
    """
    if scorer.is_session_level_scorer:
        # For conversation judges, group traces by session_id
        session_groups = _group_traces_by_session_id(trace_ids, tracking_store)
        return list(session_groups.values())
    else:
        # For single-turn judges, batch traces into fixed-size batches
        batch_size = MLFLOW_SERVER_SCORER_INVOKE_BATCH_SIZE.get()
        return [trace_ids[i : i + batch_size] for i in range(0, len(trace_ids), batch_size)]


def run_online_scoring_scheduler() -> None:
    """
    Periodic task that fetches active online scorers and submits scoring jobs.

    Groups scorers by experiment_id and submits two jobs per experiment:
    1. Trace-level scoring job for single-turn scorers
    2. Session-level scoring job for session scorers

    Groups are shuffled to prevent starvation when there are limited job runners available.
    """
    tracking_store = _get_tracking_store()
    online_scorers = tracking_store.get_active_online_scorers()
    _logger.info(f"Online scoring scheduler found {len(online_scorers)} active scorers")

    scorers_by_experiment: dict[str, list[OnlineScorer]] = defaultdict(list)
    for scorer in online_scorers:
        scorers_by_experiment[scorer.online_config.experiment_id].append(scorer)

    # Shuffle configs randomly to prevent scorer starvation when there are
    # limited job runners available
    experiment_groups = list(scorers_by_experiment.items())
    random.shuffle(experiment_groups)
    _logger.info(
        f"Grouped into {len(experiment_groups)} experiments, submitting jobs per experiment"
    )

    for experiment_id, scorers in experiment_groups:
        # Separate scorers by type
        session_level_scorers = []
        trace_level_scorers = []

        for scorer in scorers:
            try:
                scorer_obj = Scorer.model_validate_json(scorer.serialized_scorer)
                if scorer_obj.is_session_level_scorer:
                    session_level_scorers.append(scorer)
                else:
                    trace_level_scorers.append(scorer)
            except Exception as e:
                _logger.warning(
                    f"Failed to load scorer '{scorer.name}'; scorer will be skipped: {e}"
                )

        # Only submit jobs for scorer types that exist
        if trace_level_scorers:
            _logger.info(
                f"Submitting trace scoring job for experiment {experiment_id} "
                f"with {len(trace_level_scorers)} scorers"
            )
            trace_scorer_dicts = [asdict(scorer) for scorer in trace_level_scorers]
            submit_job(
                run_online_trace_scorer_job,
                {"experiment_id": experiment_id, "online_scorers": trace_scorer_dicts},
            )

        if session_level_scorers:
            _logger.info(
                f"Submitting session scoring job for experiment {experiment_id} "
                f"with {len(session_level_scorers)} scorers"
            )
            session_scorer_dicts = [asdict(scorer) for scorer in session_level_scorers]
            submit_job(
                run_online_session_scorer_job,
                {"experiment_id": experiment_id, "online_scorers": session_scorer_dicts},
            )



"""
Phoenix (Arize) integration for MLflow.
This module provides integration with Phoenix evaluators, allowing them to be used
with MLflow's scorer interface.
Example usage:
.. code-block:: python
    from mlflow.genai.scorers.phoenix import get_scorer

    scorer = get_scorer("Hallucination", model="openai:/gpt-4")
    feedback = scorer(
        inputs="What is MLflow?", outputs="MLflow is a platform...", trace=trace
    )
"""

from __future__ import annotations

import logging
from typing import Any, ClassVar

from pydantic import PrivateAttr

from mlflow.entities.assessment import Feedback
from mlflow.entities.assessment_source import AssessmentSource, AssessmentSourceType
from mlflow.entities.trace import Trace
from mlflow.genai.judges.builtin import _MODEL_API_DOC
from mlflow.genai.judges.utils import get_default_model
from mlflow.genai.scorers import FRAMEWORK_METADATA_KEY
from mlflow.genai.scorers.base import Scorer
from mlflow.genai.scorers.phoenix.models import create_phoenix_model
from mlflow.genai.scorers.phoenix.registry import get_evaluator_class
from mlflow.genai.scorers.phoenix.utils import map_scorer_inputs_to_phoenix_record
from mlflow.utils.annotations import experimental
from mlflow.utils.docstring_utils import format_docstring

_logger = logging.getLogger(__name__)


@experimental(version="3.9.0")
@format_docstring(_MODEL_API_DOC)
class PhoenixScorer(Scorer):
    """
    Base class for Phoenix metric scorers.

    Args:
        metric_name: Name of the Phoenix metric (e.g., "Hallucination")
        model: {{ model }}
    """

    _evaluator: Any = PrivateAttr()
    _model: str = PrivateAttr()

    def __init__(
        self,
        metric_name: str | None = None,
        model: str | None = None,
        **evaluator_kwargs: Any,
    ):
        if metric_name is None:
            metric_name = self.metric_name
        super().__init__(name=metric_name)
        model = model or get_default_model()
        self._model = model
        phoenix_model = create_phoenix_model(model)
        evaluator_class = get_evaluator_class(metric_name)
        self._evaluator = evaluator_class(model=phoenix_model, **evaluator_kwargs)

    def __call__(
        self,
        *,
        inputs: Any = None,
        outputs: Any = None,
        expectations: dict[str, Any] | None = None,
        trace: Trace | None = None,
    ) -> Feedback:
        """
        Evaluate using the wrapped Phoenix evaluator.

        Args:
            inputs: The input to evaluate
            outputs: The output to evaluate
            expectations: Expected values and context for evaluation
            trace: MLflow trace for evaluation
        Returns:
            Feedback object with score, rationale, and metadata
        """
        assessment_source = AssessmentSource(
            source_type=AssessmentSourceType.LLM_JUDGE,
            source_id=self._model,
        )
        try:
            record = map_scorer_inputs_to_phoenix_record(
                inputs=inputs,
                outputs=outputs,
                expectations=expectations,
                trace=trace,
            )
            label, score, explanation = self._evaluator.evaluate(record=record)
            return Feedback(
                name=self.name,
                value=label,
                rationale=explanation or None,
                source=assessment_source,
                metadata={
                    FRAMEWORK_METADATA_KEY: "phoenix",
                    "score": score,
                },
            )
        except Exception as e:
            _logger.error(f"Error evaluating Phoenix metric {self.name}: {e}")
            return Feedback(
                name=self.name,
                error=e,
                source=assessment_source,
            )


@experimental(version="3.9.0")
@format_docstring(_MODEL_API_DOC)
def get_scorer(
    metric_name: str,
    model: str | None = None,
    **evaluator_kwargs: Any,
) -> PhoenixScorer:
    """
    Get a Phoenix metric as an MLflow scorer.

    Args:
        metric_name: Name of the Phoenix metric (e.g., "Hallucination", "Relevance")
        model: {{ model }}
        evaluator_kwargs: Additional keyword arguments to pass to the Phoenix evaluator.

    Returns:
        PhoenixScorer instance that can be called with MLflow's scorer interface
    Examples:
        >>> scorer = get_scorer("Hallucination", model="openai:/gpt-4")
        >>> feedback = scorer(
        ...     inputs="What is MLflow?",
        ...     outputs="MLflow is a platform...",
        ...     expectations={"context": "MLflow is an ML platform."},
        ... )
        >>> scorer = get_scorer("Relevance", model="databricks")
        >>> feedback = scorer(trace=trace)
    """
    return PhoenixScorer(
        metric_name=metric_name,
        model=model,
        **evaluator_kwargs,
    )


@experimental(version="3.9.0")
class Hallucination(PhoenixScorer):
    """
    Detects hallucinations where the LLM fabricates information not present in the context.

    Args:
        model: Model URI (e.g., "openai:/gpt-4", "databricks", "databricks:/endpoint")

    Examples:
        .. code-block:: python
            from mlflow.genai.scorers.phoenix import Hallucination

            scorer = Hallucination(model="openai:/gpt-4")
            feedback = scorer(
                inputs="What is the capital of France?",
                outputs="Paris is the capital of France.",
                expectations={"context": "France is in Europe. Its capital is Paris."},
            )
    """

    metric_name: ClassVar[str] = "Hallucination"


@experimental(version="3.9.0")
class Relevance(PhoenixScorer):
    """
    Evaluates whether the retrieved context is relevant to the input query.

    Args:
        model: Model URI (e.g., "openai:/gpt-4", "databricks", "databricks:/endpoint")

    Examples:
        .. code-block:: python
            from mlflow.genai.scorers.phoenix import Relevance

            scorer = Relevance(model="databricks")
            feedback = scorer(
                inputs="What is machine learning?",
                expectations={"context": "ML is a subset of AI..."},
            )
    """

    metric_name: ClassVar[str] = "Relevance"


@experimental(version="3.9.0")
class Toxicity(PhoenixScorer):
    """
    Detects toxic content in the model's response.

    Args:
        model: Model URI (e.g., "openai:/gpt-4", "databricks", "databricks:/endpoint")

    Examples:
        .. code-block:: python
            from mlflow.genai.scorers.phoenix import Toxicity

            scorer = Toxicity()
            feedback = scorer(outputs="This is a friendly response.")
    """

    metric_name: ClassVar[str] = "Toxicity"


@experimental(version="3.9.0")
class QA(PhoenixScorer):
    """
    Evaluates whether the answer correctly addresses the question based on reference.

    Args:
        model: Model URI (e.g., "openai:/gpt-4", "databricks", "databricks:/endpoint")

    Examples:
        .. code-block:: python
            from mlflow.genai.scorers.phoenix import QA

            scorer = QA(model="openai:/gpt-4o")
            feedback = scorer(
                inputs="What is 2+2?",
                outputs="4",
                expectations={"context": "Basic arithmetic: 2+2=4"},
            )
    """

    metric_name: ClassVar[str] = "QA"


@experimental(version="3.9.0")
class Summarization(PhoenixScorer):
    """
    Evaluates the quality of a summarization against the original document.

    Args:
        model: Model URI (e.g., "openai:/gpt-4", "databricks", "databricks:/endpoint")

    Examples:
        .. code-block:: python
            from mlflow.genai.scorers.phoenix import Summarization

            scorer = Summarization()
            feedback = scorer(
                inputs="Long document text...",
                outputs="Brief summary of the document.",
            )
    """

    metric_name: ClassVar[str] = "Summarization"


__all__ = [
    # Core classes
    "PhoenixScorer",
    "get_scorer",
    # Metric scorers
    "Hallucination",
    "Relevance",
    "Toxicity",
    "QA",
    "Summarization",
]



"""add online_scoring_configs table

Create Date: 2025-01-27 12:00:00.000000

"""

import sqlalchemy as sa
from alembic import op

from mlflow.store.tracking.dbmodels.models import SqlOnlineScoringConfig

# revision identifiers, used by Alembic.
revision = "2c33131f4dae"
down_revision = "c9d4e5f6a7b8"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        SqlOnlineScoringConfig.__tablename__,
        sa.Column("online_scoring_config_id", sa.String(length=36), nullable=False),
        sa.Column("scorer_id", sa.String(length=36), nullable=False),
        sa.Column("sample_rate", sa.types.Float(precision=53), nullable=False),
        sa.Column("experiment_id", sa.Integer(), nullable=False),
        sa.Column("filter_string", sa.Text(), nullable=True),
        sa.ForeignKeyConstraint(
            ["scorer_id"],
            ["scorers.scorer_id"],
            name="fk_online_scoring_configs_scorer_id",
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["experiment_id"],
            ["experiments.experiment_id"],
            name="fk_online_scoring_configs_experiment_id",
        ),
        sa.PrimaryKeyConstraint("online_scoring_config_id", name="online_scoring_config_pk"),
    )


def downgrade():
    op.drop_table(SqlOnlineScoringConfig.__tablename__)



import logging
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from mlflow.genai.utils.type import FunctionCall
    from mlflow.types.chat import ChatTool

_logger = logging.getLogger(__name__)


def format_available_tools(available_tools: list["ChatTool"]) -> str:
    """Format available tools with descriptions and parameters.

    Args:
        available_tools: The set of available tools

    Returns:
        Formatted string representation of available tools

    Example:
        >>> # Output format:
        >>> # - search: Search for information on the web
        >>> #     - query (required): string - The search query to execute
        >>> #     - max_results (optional): integer - Maximum number of results
        >>> # - translate: Translate text to another language
        >>> #     - text (required): string - The text to translate
        >>> #     - target (required): string - Target language code
    """
    available_tools_parts = []
    for tool in available_tools:
        if not tool.function:
            _logger.warning(f"Skipping tool with missing function definition: {tool}")
            continue

        tool_str = f"- {tool.function.name}"
        if tool.function.description:
            tool_str += f": {tool.function.description}"

        if tool.function.parameters and tool.function.parameters.properties:
            params = tool.function.parameters
            required_params = set(params.required or [])
            param_lines = []

            for param_name, param_prop in params.properties.items():
                is_required = param_name in required_params
                required_marker = " (required)" if is_required else " (optional)"
                param_line = f"    - {param_name}{required_marker}"

                if hasattr(param_prop, "type") and param_prop.type:
                    param_line += f": {param_prop.type}"

                if param_prop.description:
                    param_line += f" - {param_prop.description}"

                param_lines.append(param_line)

            if param_lines:
                tool_str += "\n" + "\n".join(param_lines)

        available_tools_parts.append(tool_str)
    return "\n\n".join(available_tools_parts) if available_tools_parts else "No tools available"


def format_tools_called(tools_called: list["FunctionCall"]) -> str:
    """Format tools called with step numbers, arguments, and outputs.

    Args:
        tools_called: The sequence of tools that were called by the agent.
            Each element should be a FunctionCall object.

    Returns:
        Formatted string representation of tools called

    Example:
        >>> # Output format:
        >>> # Tool Call 1: search
        >>> #   Input Arguments: {"query": "capital of France"}
        >>> #   Output: Paris
        >>> #
        >>> # Tool Call 2: translate
        >>> #   Input Arguments: {"text": "Paris", "target": "es"}
        >>> #   Output: París
    """
    tools_called_parts = []
    for idx, tool in enumerate(tools_called, start=1):
        tool_name = tool.name
        tool_args = tool.arguments or {}
        tool_output = tool.outputs or "(no output)"

        tool_str = f"Tool Call {idx}: {tool_name}\n"
        tool_str += f"  Input Arguments: {tool_args}\n"
        tool_str += f"  Output: {tool_output}"
        if tool.exception:
            tool_str += f"\n  Exception: {tool.exception}"
        tools_called_parts.append(tool_str)
    return "\n\n".join(tools_called_parts) if tools_called_parts else "No tools called"



from __future__ import annotations

from mlflow.exceptions import MlflowException
from mlflow.genai.scorers.phoenix.utils import check_phoenix_installed

_METRIC_REGISTRY = {
    "Hallucination": "HallucinationEvaluator",
    "Relevance": "RelevanceEvaluator",
    "Toxicity": "ToxicityEvaluator",
    "QA": "QAEvaluator",
    "Summarization": "SummarizationEvaluator",
}


def get_evaluator_class(metric_name: str):
    """
    Get the Phoenix evaluator class for a given metric name.

    Args:
        metric_name: Name of the metric (e.g., "Hallucination")

    Returns:
        The Phoenix evaluator class

    Raises:
        MlflowException: If the metric is not supported
    """
    check_phoenix_installed()

    if metric_name not in _METRIC_REGISTRY:
        available_metrics = ", ".join(sorted(_METRIC_REGISTRY.keys()))
        raise MlflowException.invalid_parameter_value(
            f"Unknown Phoenix metric: '{metric_name}'. Available metrics: {available_metrics}"
        )

    import phoenix.evals as phoenix_evals

    evaluator_class_name = _METRIC_REGISTRY[metric_name]
    return getattr(phoenix_evals, evaluator_class_name)



"""Constants for online scoring."""

from mlflow.tracing.constant import TraceMetadataKey

# Maximum lookback period to prevent getting stuck on old failing traces (1 hour)
MAX_LOOKBACK_MS = 60 * 60 * 1000

# Buffer time to wait before considering a session complete (5 minutes)
# Sessions with no new traces for this duration are considered complete
SESSION_COMPLETION_BUFFER_MS = 5 * 60 * 1000

# Maximum traces to include in a single scoring job
MAX_TRACES_PER_JOB = 500

# Maximum sessions to include in a single scoring job
MAX_SESSIONS_PER_JOB = 100

# Filter to exclude eval run traces (traces generated from MLflow runs)
EXCLUDE_EVAL_RUN_TRACES_FILTER = f"metadata.{TraceMetadataKey.SOURCE_RUN} IS NULL"



from mlflow.assistant.providers.base import AssistantProvider
from mlflow.assistant.providers.claude_code import ClaudeCodeProvider

__all__ = ["AssistantProvider", "ClaudeCodeProvider", "list_providers"]

# Registry of all available providers
_PROVIDERS: list[type[AssistantProvider]] = [
    ClaudeCodeProvider,
]


def list_providers() -> list[AssistantProvider]:
    """Return instances of all available providers."""
    return [provider() for provider in _PROVIDERS]



from mlflow.genai.optimize.optimizers.base import BasePromptOptimizer
from mlflow.genai.optimize.optimizers.gepa_optimizer import GepaPromptOptimizer
from mlflow.genai.optimize.optimizers.metaprompt_optimizer import MetaPromptOptimizer

__all__ = ["BasePromptOptimizer", "GepaPromptOptimizer", "MetaPromptOptimizer"]



"""Online scoring subpackage for scheduled scorer execution."""

from mlflow.genai.scorers.online.entities import (
    CompletedSession,
    OnlineScorer,
    OnlineScoringConfig,
)
from mlflow.genai.scorers.online.sampler import OnlineScorerSampler
from mlflow.genai.scorers.online.session_checkpointer import OnlineSessionCheckpointManager
from mlflow.genai.scorers.online.session_processor import OnlineSessionScoringProcessor
from mlflow.genai.scorers.online.trace_checkpointer import OnlineTraceCheckpointManager
from mlflow.genai.scorers.online.trace_loader import OnlineTraceLoader
from mlflow.genai.scorers.online.trace_processor import OnlineTraceScoringProcessor

__all__ = [
    "CompletedSession",
    "OnlineScorer",
    "OnlineScorerSampler",
    "OnlineScoringConfig",
    "OnlineSessionCheckpointManager",
    "OnlineSessionScoringProcessor",
    "OnlineTraceCheckpointManager",
    "OnlineTraceLoader",
    "OnlineTraceScoringProcessor",
]



CONVERSATIONAL_GUIDELINES_ASSESSMENT_NAME = "conversational_guidelines"

CONVERSATIONAL_GUIDELINES_PROMPT = """\
Consider the following conversation history between a user and an assistant.
Your task is to evaluate whether the assistant's responses throughout the conversation comply with
the provided guidelines and output exactly one label: "yes" or "no".

<guidelines>
{{ guidelines }}
</guidelines>

Evaluation criteria:
- Assess whether EVERY assistant response in the conversation follows ALL the provided guidelines.
- Focus on judging only the assistant's responses, not the user's messages.
- Only focus on the provided guidelines and not the correctness, relevance, or effectiveness of the responses.
- A guideline violation at ANY point in the conversation means the entire conversation fails.
- If none of the guidelines apply to the given conversation, the result must be "yes".

Output "yes" if all assistant responses comply with all guidelines throughout the entire conversation.
Output "no" if any assistant response violates any guideline at any point in the conversation.

<conversation>{{ conversation }}</conversation>
"""  # noqa: E501



from typing import TYPE_CHECKING

from mlflow.genai.judges.utils.formatting_utils import (
    format_available_tools,
    format_tools_called,
)
from mlflow.genai.prompts.utils import format_prompt

if TYPE_CHECKING:
    from mlflow.genai.utils.type import FunctionCall
    from mlflow.types.chat import ChatTool

# NB: User-facing name for the is_tool_call_efficient assessment.
TOOL_CALL_EFFICIENCY_FEEDBACK_NAME = "tool_call_efficiency"

TOOL_CALL_EFFICIENCY_PROMPT_INSTRUCTIONS = """\
Consider the agent's tool usage for redundancy and inefficiency.

Given the user's request, the available tools, and the sequence of tools called by the agent, \
determine whether any tool calls were unnecessary or could have been made more efficient. In your \
analysis, treat retries caused by temporary tool failures (e.g., timeouts, transient errors) as \
efficient and not redundant.

Consider in particular:

Calls to the same tool with identical or very similar arguments

Repeated calls to the same tool with the same parameters

Multiple calls that could reasonably have been consolidated into a single call

<request>
{{request}}
</request>

<available_tools>
{{available_tools}}
</available_tools>

<tools_called>
{{tools_called}}
</tools_called>"""

TOOL_CALL_EFFICIENCY_PROMPT_OUTPUT = """

Please evaluate whether the agent's tool usage is efficient and free of redundancy using only the following json format. Return "yes" if the tool usage is efficient and free of redundancy, otherwise return "no".
Do not use any markdown formatting or output additional lines.
{
  "rationale": "Reason for the assessment. If redundant tool calls are found, identify which specific calls are redundant and explain why. If no redundancy is found, explain why the tool usage is efficient. Start each rationale with `Let's think step by step`",
  "result": "yes|no"
}\
"""  # noqa: E501

TOOL_CALL_EFFICIENCY_PROMPT = (
    TOOL_CALL_EFFICIENCY_PROMPT_INSTRUCTIONS + TOOL_CALL_EFFICIENCY_PROMPT_OUTPUT
)


def get_prompt(
    request: str,
    tools_called: list["FunctionCall"],
    available_tools: list["ChatTool"],
) -> str:
    """Generate tool call efficiency evaluation prompt.

    Args:
        request: The original user request that the agent is trying to fulfill
        tools_called: The sequence of tools that were called by the agent.
            Each element should be a FunctionCall object.
        available_tools: The set of available tools

    Returns:
        Formatted prompt string
    """
    available_tools_str = format_available_tools(available_tools)
    tools_called_str = format_tools_called(tools_called)

    return format_prompt(
        TOOL_CALL_EFFICIENCY_PROMPT,
        request=request,
        available_tools=available_tools_str,
        tools_called=tools_called_str,
    )



import json
from typing import TYPE_CHECKING

from mlflow.genai.judges.utils.formatting_utils import (
    format_available_tools,
    format_tools_called,
)
from mlflow.genai.prompts.utils import format_prompt

if TYPE_CHECKING:
    from mlflow.genai.utils.type import FunctionCall
    from mlflow.types.chat import ChatTool

TOOL_CALL_CORRECTNESS_FEEDBACK_NAME = "tool_call_correctness"

# Shared output format for all prompt variants
_OUTPUT_FORMAT = """
Please evaluate whether the agent's tool calls and their arguments are correct and reasonable using only the following json format. Return "yes" if the tool calls and arguments are correct and reasonable, otherwise return "no".
Do not use any markdown formatting or output additional lines.
{
  "rationale": "Reason for the assessment. If incorrect or unreasonable tool calls are found, identify which specific calls or arguments are problematic and explain why. If all tool calls and arguments are correct, explain why they are appropriate. Start each rationale with `Let's think step by step`",
  "result": "yes|no"
}\
"""  # noqa: E501

# Ordering instruction variants
ORDERING_INSTRUCTION_CHECK = (
    "3) Ordering\n- Consider whether the order of tool calls matches the expected order."
)
ORDERING_INSTRUCTION_IGNORE = (
    "Note: The order of tool calls does not need to match. You should not penalize the agent for "
    "calling tools in a different order than the expected order."
)

# Evaluation criteria for ground-truth-free mode
_GROUND_TRUTH_FREE_CRITERIA = """\
1) Need for tools
- Was using any tool necessary or helpful for this request?
- Did the agent fail to use an obviously appropriate tool that was available?

2) Tool selection
- For each step, is the chosen tool a good match for the subtask, given the tool descriptions?
- Did the agent avoid tools that are clearly irrelevant, overpowered, or disallowed for the request?

3) Arguments and intent alignment
- Do the arguments match the tool's schema?
- Are the arguments clearly grounded in the user's request and the tool's documented purpose?
- Across calls, are key parameters provided in ways that logically follow from prior tool outputs
or user messages, rather than arbitrary changes?

4) Tool flow and combinations
- When multiple tools are used, is the overall sequence of tool choices logically sound?
- Are follow-up tool calls justified by what the agent appears to be trying to achieve with respect
to the user's request?

{{ordering_instruction}}"""

# Evaluation criteria for full expectations (names + arguments)
_FULL_EXPECTATIONS_CRITERIA = """\
1) Tool selection match
- Are the same tools being called (or semantically equivalent alternatives)?
- Are there any missing or extra tool calls compared to expectations?

2) Argument match
- Do the actual arguments convey the same intent as the expected arguments, even if phrased
differently?
- Are there any significant differences in argument values that would change the outcome?

{{ordering_instruction}}"""

# Evaluation criteria for partial expectations (names only)
_PARTIAL_EXPECTATIONS_CRITERIA = """\
1) Tool selection match
- Do the actual tool calls match the expected tool names?
- Are there any missing or extra tools compared to expectations?

2) Argument reasonableness
- Are the arguments provided reasonable given the user request and tool definitions?
- Do the arguments match the tool's schema and documented purpose?

{{ordering_instruction}}"""

# Unified prompt template
_PROMPT_TEMPLATE = """\
{{preamble}}

Focus only on the choice of tools and the arguments passed to them. Do NOT judge whether the tools'
outputs or implementations are correct.

Evaluate:

{{evaluation_criteria}}

<request>
{{request}}
</request>

{{expected_section}}<actual_tool_calls>
{{tools_called}}
</actual_tool_calls>

<available_tools>
{{available_tools}}
</available_tools>"""

# Preamble variants
_GROUND_TRUTH_FREE_PREAMBLE = """\
Consider whether the agent selected appropriate tools and called with the correct arguments for the
task.

Given the user's request, the available tools (including their described capabilities/constraints),
and the sequence of tool calls made by the agent, evaluate if the agent chose suitable tools and
used them in a reasonable way."""

_FULL_EXPECTATIONS_PREAMBLE = """\
Compare the actual tool calls against the expected tool calls to determine if they are correct.

Given the user's request, the expected tool calls (ground truth), and the actual tool calls made
by the agent, evaluate whether the actual tool calls semantically match the expected ones."""

_PARTIAL_EXPECTATIONS_PREAMBLE = """\
Evaluate tool call correctness by comparing tool selection against expected tool names, and
evaluating whether arguments are reasonable.

Given the user's request, the expected tool names (ground truth), the available tools (including
their described capabilities/constraints), and the actual tool calls made by the agent, evaluate
whether the agent selected the correct tools and used reasonable arguments.\
"""

# Used by ToolCallCorrectness.instructions property for serialization
TOOL_CALL_CORRECTNESS_PROMPT_INSTRUCTIONS = (
    _GROUND_TRUTH_FREE_PREAMBLE
    + "\n\nFocus only on the choice of tools and the arguments passed to them. Do NOT judge "
    "whether the tools'\noutputs or implementations are correct.\n\nEvaluate:\n\n"
    + _GROUND_TRUTH_FREE_CRITERIA
    + "\n\n<request>\n{{request}}\n</request>\n\n<available_tools>\n{{available_tools}}\n"
    "</available_tools>\n\n<tools_called>\n{{tools_called}}\n</tools_called>"
)


def _format_expected_calls(expected_calls: list["FunctionCall"], include_arguments: bool) -> str:
    lines = []
    for i, call in enumerate(expected_calls, 1):
        if include_arguments:
            lines.append(f"Expected Tool Call {i}: {call.name}")
            if call.arguments:
                lines.append(f"  Arguments: {json.dumps(call.arguments)}")
            else:
                lines.append("  Arguments: empty")
        else:
            lines.append(f"Expected Tool {i}: {call.name}")
    return "\n".join(lines) if lines else "No expected tool calls provided."


def get_prompt(
    request: str,
    tools_called: list["FunctionCall"],
    available_tools: list["ChatTool"],
    expected_calls: list["FunctionCall"] | None = None,
    include_arguments: bool = True,
    check_order: bool = False,
) -> str:
    """
    Generate tool call correctness evaluation prompt.

    Args:
        request: The original user request that the agent is trying to fulfill
        tools_called: The sequence of tools that were called by the agent
        available_tools: The set of available tools
        expected_calls: Optional list of expected tool calls for ground-truth comparison.
            If None, uses ground-truth-free evaluation.
        include_arguments: If True, compare both tool names and arguments (full expectations).
            If False, compare only tool names (partial expectations).
        check_order: If True, ask LLM to consider ordering of tool calls.
    """
    available_tools_str = format_available_tools(available_tools)
    tools_called_str = format_tools_called(tools_called)
    ordering = ORDERING_INSTRUCTION_CHECK if check_order else ORDERING_INSTRUCTION_IGNORE

    if expected_calls is None:
        preamble = _GROUND_TRUTH_FREE_PREAMBLE
        criteria = _GROUND_TRUTH_FREE_CRITERIA.replace("{{ordering_instruction}}", ordering)
        expected_section = ""
    elif include_arguments:
        preamble = _FULL_EXPECTATIONS_PREAMBLE
        criteria = _FULL_EXPECTATIONS_CRITERIA.replace("{{ordering_instruction}}", ordering)
        expected_calls_str = _format_expected_calls(expected_calls, include_arguments)
        expected_section = (
            f"<expected_tool_calls>\n{expected_calls_str}\n</expected_tool_calls>\n\n"
        )
    else:
        preamble = _PARTIAL_EXPECTATIONS_PREAMBLE
        criteria = _PARTIAL_EXPECTATIONS_CRITERIA.replace("{{ordering_instruction}}", ordering)
        expected_calls_str = _format_expected_calls(expected_calls, include_arguments)
        expected_section = (
            f"<expected_tool_calls>\n{expected_calls_str}\n</expected_tool_calls>\n\n"
        )

    return format_prompt(
        _PROMPT_TEMPLATE + _OUTPUT_FORMAT,
        preamble=preamble,
        evaluation_criteria=criteria,
        expected_section=expected_section,
        request=request,
        available_tools=available_tools_str,
        tools_called=tools_called_str,
    )



"""MLflow GenAI Judge Optimizers."""

from mlflow.genai.judges.optimizers.memalign import MemAlignOptimizer
from mlflow.genai.judges.optimizers.simba import SIMBAAlignmentOptimizer

__all__ = [
    "MemAlignOptimizer",
    "SIMBAAlignmentOptimizer",
]
